# -*- encoding: utf-8 -*-
# Software Management Providers
#
# Copyright (C) 2012 Red Hat, Inc.  All rights reserved.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#
# Authors: Jan Safranek <jsafrane@redhat.com>
"""
Pylint checker for OpenLMI storage modules and classes.
"""

import re
from pylint.interfaces import IASTNGChecker
from pylint.checkers import BaseChecker

import pdb

class LMIStorageChecker(BaseChecker):
    """
    Add/remove checks specific for openlmi.storage
    """
    name = 'lmi_storage'
    msgs = { 'W9922': ('Dummy', "This is a dummy message.")}
    __implements__ = IASTNGChecker

    # this is important so that your checker is executed before others
    priority = -2

    def visit_module(self, node):
        """
        Suppress R0201: Method could be a function
        """
        if node.name.startswith("openlmi.storage"):
            self.linter.disable('R0201')

def register(linter):
    """required method to auto register our checker"""
    linter.register_checker(LMIStorageChecker(linter))



from setuptools import setup
from lmi.software import __version__

setup(
    name='openlmi-software',
    description='Software Management providers',
    author='Michal Minar',
    author_email='miminar@redhat.com',
    url='https://fedorahosted.org/openlmi/',
    version=__version__,
    namespace_packages=['lmi'],
    packages=[
        'lmi',
        'lmi.software',
        'lmi.software.core',
        'lmi.software.util',
        'lmi.software.yumdb'],
    data_files=[('/etc/openlmi/software',
            ['config/software.conf', 'config/yum_worker_logging.conf'])],
    install_requires=['openlmi'],
    license="LGPLv2+",
    classifiers=[
        'License :: OSI Approved :: GNU Lesser General Public License v2 or later (LGPLv2+)',
        'Operating System :: POSIX :: Linux',
        'Topic :: System :: Systems Administration',
        ]
    )



#!/usr/bin/python
#
# Copyright (C) 2013 Red Hat, Inc.  All rights reserved.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#
# Authors: Jan Synacek <jsynacek@redhat.com>

from test_base import LogicalFileTestBase
import unittest
import pywbem
import os
import stat
import shutil
import subprocess

class TestLogicalFile(LogicalFileTestBase):
    """
    Exhaustive LogicalFile tests.
    """

    def setUp(self):
        super(TestLogicalFile, self).setUp()
        self.files = {'data':{'path' : self.testdir + "/data",
                              'class': 'LMI_DataFile',
                              'props': {'Readable':True,
                                        'Writeable':False,
                                        'Executable':True,
                                        'SELinuxCurrentContext':'testlf_u:testlf_r:testlf_data_t:s0',
                                        'SELinuxExpectedContext':''}},
                      'dir':{'path' : self.testdir + "/dir",
                             'class': 'LMI_UnixDirectory',
                             'props': {}},
                      'hardlink':{'path' : self.testdir + "/hardlink",
                                  'class': 'LMI_DataFile',
                                  'props': {'Readable':True,
                                            'Writeable':False,
                                            'Executable':True}},
                      'symlink':{'path' : self.testdir + "/symlink",
                                 'class': 'LMI_SymbolicLink',
                                 'props': {'SELinuxCurrentContext':'testlf_u:testlf_r:testlf_symlink_t:s0',
                                           'SELinuxExpectedContext':''}},
                      'fifo':{'path' : self.testdir + "/fifo",
                              'class': 'LMI_FIFOPipeFile',
                              'props': {}},
                      'chdev':{'path' : self.testdir + "/chdev",
                               'class': 'LMI_UnixDeviceFile',
                               'props': {'DeviceMajor':2, 'DeviceMinor':4,
                                         'DeviceFileType':3}},
                      'bldev':{'path' : self.testdir + "/bldev",
                               'class': 'LMI_UnixDeviceFile',
                               'props': {'DeviceMajor':3, 'DeviceMinor':5,
                                         'DeviceFileType':2}},
                      '..':{'path' : os.path.realpath(self.testdir + "/.."),
                            'class': 'LMI_UnixDirectory',
                            'props': {}}}

        self.transient_file = {'path' : self.testdir + "/transient",
                               'class' : 'LMI_UnixDirectory',
                               'props' : {'FSCreationClassName' : 'LMI_TransientFileSystem',
                                          'FSName' : 'PATH=' + self.testdir + "/transient"}}

        self.num_files = len(self.files.keys()) + 1

        self.cop = pywbem.CIMInstanceName(classname='LMI_UnixDirectory',
                                          namespace='root/cimv2',
                                          keybindings={
                                              'CSCreationClassName':self.system_cs_name,
                                              'CSName':self.SYSTEM_NAME,
                                              'FSCreationClassName':'LMI_LocalFileSystem',
                                              'FSName':self.fsname,
                                              'CreationClassName':'LMI_UnixDirectory',
                                              'Name':self.testdir
                                          })
        self._prepare()

    def tearDown(self):
        self._cleanup()

    def _prepare(self):
        try:
            os.stat(self.testdir)
        except OSError:
            os.makedirs(self.testdir)
            data_file = self.files['data']['path']
            data_props = self.files['data']['props']
            f = open(data_file, "w+")
            f.write("hello")
            f.close()
            os.chmod(data_file, 0550)
            if self.selinux_enabled:
                labels = data_props['SELinuxCurrentContext'].split(':')
                out = subprocess.check_output(['chcon', '-h',
                                               '-u', labels[0],
                                               '-r', labels[1],
                                               '-t', labels[2], data_file])
                out = subprocess.check_output(['matchpathcon', '-n', data_file])
                data_props['SELinuxExpectedContext'] = out[:-1] # remove \n
            os.mkdir(self.files['dir']['path'])
            os.link(data_file, self.files['hardlink']['path'])
            slink_file = self.files['symlink']['path']
            slink_props = self.files['symlink']['props']
            os.symlink(data_file, slink_file)
            if self.selinux_enabled:
                labels = slink_props['SELinuxCurrentContext'].split(':')
                out = subprocess.check_output(['chcon', '-h',
                                               '-u', labels[0],
                                               '-r', labels[1],
                                               '-t', labels[2], slink_file])
                out = subprocess.check_output(['matchpathcon', '-n', slink_file])
                slink_props['SELinuxExpectedContext'] = out[:-1] # remove \n
            os.mkfifo(self.files['fifo']['path'])
            chdev = self.files['chdev']
            chdev_device = os.makedev(chdev['props']['DeviceMajor'], chdev['props']['DeviceMinor'])
            os.mknod(chdev['path'], 0666 | stat.S_IFCHR, chdev_device)
            bldev = self.files['bldev']
            bldev_device = os.makedev(bldev['props']['DeviceMajor'], bldev['props']['DeviceMinor'])
            os.mknod(bldev['path'], 0666 | stat.S_IFBLK, bldev_device)
            transient_file = self.transient_file['path']
            os.mkdir(transient_file)
            subprocess.call(['mount', '-t', 'tmpfs', 'tmpfs', transient_file, '-o', 'size=1M'])

    def _cleanup(self):
        subprocess.call(['umount', self.transient_file['path']])
        shutil.rmtree(self.testdir)

    def test_lmi_directorycontainsfile(self):
        assoc_class = 'LMI_DirectoryContainsFile'
        ### Associators and AssociatorNames
        for assoc_method in [self.wbemconnection.Associators, self.wbemconnection.AssociatorNames]:
            assocs = assoc_method(self.cop, AssocClass=assoc_class)
            self.assertEquals(len(assocs), self.num_files)
            for k, f in self.files.iteritems():
                # test that the files are actually there and have the correct class name
                match = filter(lambda a: a['Name'] == f['path'], assocs)
                self.assertEquals(len(match), 1)
                self.assertEquals(match[0].classname, f['class'])
                if not isinstance(match[0], pywbem.CIMInstanceName):
                    # test some selected properties
                    if k == 'data' or k == 'hardlink':
                        self.assertEquals(match[0]['Readable'], f['props']['Readable'])
                        self.assertEquals(match[0]['Writeable'], f['props']['Writeable'])
                        self.assertEquals(match[0]['Executable'], f['props']['Executable'])
                    if k == 'chdev' or k == 'bldev':
                        self.assertEqual(match[0]['DeviceMajor'], str(f['props']['DeviceMajor']))
                        self.assertEqual(match[0]['DeviceMinor'], str(f['props']['DeviceMinor']))
                        self.assertEqual(match[0]['DeviceFileType'], f['props']['DeviceFileType'])
                    if k == 'symlink':
                        self.assertEqual(match[0]['TargetFile'], self.files['data']['path'])

                # test the other side of LMI_DirectoryContainsFile
                if k != '..':
                    cop_file = pywbem.CIMInstanceName(classname=f['class'],
                                                  namespace='root/cimv2',
                                                  keybindings={
                                                      'CSCreationClassName':self.system_cs_name,
                                                      'CSName':self.SYSTEM_NAME,
                                                      'FSCreationClassName':'LMI_LocalFileSystem',
                                                      'FSName':self.fsname,
                                                      'CreationClassName':f['class'],
                                                      'Name':f['path']
                                                  })
                    assocs_file = assoc_method(cop_file,
                                               AssocClass=assoc_class,
                                               Role='PartComponent',
                                               ResultRole='GroupComponent',
                                               ResultClass=self.files['..']['class'])
                    self.assertEquals(len(assocs_file), 1)
                    self.assertEquals(assocs_file[0].classname, 'LMI_UnixDirectory')
                    self.assertEquals(assocs_file[0]['Name'], self.testdir)
                    # wrong Role
                    assocs_file = assoc_method(cop_file, AssocClass=assoc_class, Role='GroupComponent')
                    self.assertEquals(assocs_file, [])
                    # wrong ResultRole
                    assocs_file = assoc_method(cop_file, AssocClass=assoc_class, ResultRole='PartComponent')
                    self.assertEquals(assocs_file, [])
                    # good role and ResultRole, wrong ResultClass
                    assocs_file = assoc_method(cop_file,
                                               AssocClass=assoc_class,
                                               Role='PartComponent',
                                               ResultRole='GroupComponent',
                                               ResultClass=self.system_cs_name)
                    self.assertEquals(assocs_file, [])

        ### References and ReferenceNames
        for assoc_method in [self.wbemconnection.References, self.wbemconnection.ReferenceNames]:
            assocs = assoc_method(self.cop, ResultClass='LMI_DirectoryContainsFile')
            self.assertEquals(len(assocs), self.num_files)
            for k, f in self.files.iteritems():
                # test that the files are actually there and have the correct class name
                match = filter(lambda a: a['PartComponent']['Name'] == f['path'], assocs)
                self.assertEquals(len(match), 1)
                match_name = match[0]['PartComponent']
                self.assertEquals(match_name.classname, f['class'])
                # test the other side of LMI_DirectoryContainsFile
                if k != '..' and k != 'dir':
                    cop_file = pywbem.CIMInstanceName(classname=f['class'],
                                                  namespace='root/cimv2',
                                                  keybindings={
                                                      'CSCreationClassName':self.system_cs_name,
                                                      'CSName':self.SYSTEM_NAME,
                                                      'FSCreationClassName':'LMI_LocalFileSystem',
                                                      'FSName':self.fsname,
                                                      'CreationClassName':f['class'],
                                                      'Name':f['path']
                                                  })
                    assocs_file = assoc_method(cop_file, ResultClass='LMI_DirectoryContainsFile')
                    self.assertEquals(len(assocs_file), 1)
                    file_name = assocs_file[0]['GroupComponent']
                    self.assertEquals(file_name.classname, 'LMI_UnixDirectory')
                    self.assertEquals(file_name['Name'], self.testdir)

    def test_lmi_fileidentity(self):
        ### Associators and AssociatorNames
        assoc_class = 'LMI_FileIdentity'
        for assoc_method in [self.wbemconnection.Associators, self.wbemconnection.AssociatorNames]:
            assocs = assoc_method(self.cop, AssocClass='LMI_DirectoryContainsFile')
            self.assertEquals(len(assocs), self.num_files)
            for k, f in self.files.iteritems():
                match = filter(lambda a: a['Name'] == f['path'], assocs)
                self.assertEquals(len(match), 1)
                self.assertEquals(match[0].classname, f['class'])

                if isinstance(match[0], pywbem.CIMInstanceName):
                    match[0].path = match[0]
                ## SystemElement - LMI_UnixFile
                assocs_ident = assoc_method(match[0].path,
                                            AssocClass=assoc_class,
                                            Role='SystemElement',
                                            ResultRole='SameElement',
                                            ResultClass='LMI_UnixFile')
                self.assertEquals(len(assocs_ident), 1)
                self.assertEquals(assocs_ident[0]['LFName'], f['path'])
                self.assertEquals(assocs_ident[0]['LFCreationClassName'], f['class'])
                if not isinstance(match[0], pywbem.CIMInstanceName):
                    if self.selinux_enabled and (k == 'data' or k == 'symlink'):
                        self.assertEquals(assocs_ident[0]['SELinuxCurrentContext'],
                                          f['props']['SELinuxCurrentContext'])
                        self.assertEquals(assocs_ident[0]['SELinuxExpectedContext'],
                                          f['props']['SELinuxExpectedContext'])
                # wrong Role
                assocs_ident = assoc_method(match[0].path, AssocClass=assoc_class, Role='SameElement')
                self.assertEquals(assocs_ident, [])
                # wrong ResultRole
                assocs_ident = assoc_method(match[0].path, AssocClass=assoc_class, ResultRole='SystemElement')
                self.assertEquals(assocs_ident, [])
                # good role and ResultRole, wrong ResultClass
                assocs_ident = assoc_method(match[0].path,
                                      AssocClass=assoc_class,
                                      Role='SystemElement',
                                      ResultRole='SameElement',
                                      ResultClass=self.system_cs_name)
                self.assertEquals(assocs_ident, [])

                ## SameElement - CIM_LogicalFile
                cop_ident = pywbem.CIMInstanceName(classname='LMI_UnixFile',
                                                   namespace='root/cimv2',
                                                   keybindings={
                                                       'CSCreationClassName':self.system_cs_name,
                                                       'CSName':self.SYSTEM_NAME,
                                                       'FSCreationClassName':'LMI_LocalFileSystem',
                                                       'FSName':self.fsname,
                                                       'LFCreationClassName':f['class'],
                                                       'LFName':f['path']
                                                   })
                assocs_ident = assoc_method(cop_ident,
                                            AssocClass=assoc_class,
                                            Role='SameElement',
                                            ResultRole='SystemElement',
                                            ResultClass=f['class'])
                self.assertEquals(len(assocs_ident), 1)
                self.assertEquals(assocs_ident[0]['Name'], f['path'])
                self.assertEquals(assocs_ident[0]['CreationClassName'], f['class'])
                # wrong Role
                assocs_ident = assoc_method(cop_ident, AssocClass=assoc_class, Role='SystemElement')
                self.assertEquals(assocs_ident, [])
                # wrong ResultRole
                assocs_ident = assoc_method(cop_ident, AssocClass=assoc_class, ResultRole='SameElement')
                self.assertEquals(assocs_ident, [])
                # good role and ResultRole, wrong ResultClass
                assocs_ident = assoc_method(cop_ident,
                                      AssocClass=assoc_class,
                                      Role='SameElement',
                                      ResultRole='SystemElement',
                                      ResultClass=self.system_cs_name)
                self.assertEquals(assocs_ident, [])

        ### References and ReferenceNames
        for assoc_method in [self.wbemconnection.References, self.wbemconnection.ReferenceNames]:
            assocs = assoc_method(self.cop, ResultClass='LMI_DirectoryContainsFile')
            self.assertEquals(len(assocs), self.num_files)
            for k, f in self.files.iteritems():
                match = filter(lambda a: a['PartComponent']['Name'] == f['path'], assocs)
                self.assertEquals(len(match), 1)
                match_name = match[0]['PartComponent']
                self.assertEquals(match_name['CreationClassName'], f['class'])

                ## SystemElement - LMI_UnixFile
                assocs_ident = assoc_method(match_name, ResultClass=assoc_class)
                self.assertEquals(len(assocs_ident), 1)
                ident_name = assocs_ident[0]['SameElement']
                self.assertEquals(ident_name['LFName'], f['path'])
                self.assertEquals(ident_name['LFCreationClassName'], f['class'])


                ## SameElement - CIM_LogicalFile
                cop_ident = pywbem.CIMInstanceName(classname='LMI_UnixFile',
                                                   namespace='root/cimv2',
                                                   keybindings={
                                                       'CSCreationClassName':self.system_cs_name,
                                                       'CSName':self.SYSTEM_NAME,
                                                       'FSCreationClassName':'LMI_LocalFileSystem',
                                                       'FSName':self.fsname,
                                                       'LFCreationClassName':f['class'],
                                                       'LFName':f['path']
                                                   })
                assocs_ident = assoc_method(cop_ident, ResultClass=assoc_class)
                self.assertEquals(len(assocs_ident), 1)
                ident_name = assocs_ident[0]['SystemElement']
                self.assertEquals(ident_name['Name'], f['path'])
                self.assertEquals(ident_name['CreationClassName'], f['class'])

    def test_lmi_rootdirectory(self):
        assoc_class = 'LMI_RootDirectory'
        ### Associators and AssociatorNames
        cop = self.cop.copy()
        cop.keybindings['Name'] = '/'
        for assoc_method in [self.wbemconnection.Associators, self.wbemconnection.AssociatorNames]:
            ## PartComponent - CIM_ComputerSystem
            assocs = assoc_method(cop,
                                  AssocClass=assoc_class,
                                  Role='PartComponent',
                                  ResultRole='GroupComponent',
                                  ResultClass=self.system_cs_name)
            self.assertEquals(len(assocs), 1)
            system = assocs[0]
            self.assertEquals(system['CreationClassName'], self.system_cs_name)
            self.assertEquals(system['Name'], self.SYSTEM_NAME)
            # wrong Role
            assocs = assoc_method(cop, AssocClass=assoc_class, Role='GroupComponent')
            self.assertEquals(assocs, [])
            # wrong ResultRole
            assocs = assoc_method(cop, AssocClass=assoc_class, ResultRole='PartComponent')
            self.assertEquals(assocs, [])
            # good role and ResultRole, wrong ResultClass
            assocs = assoc_method(cop,
                                  AssocClass=assoc_class,
                                  Role='PartComponent',
                                  ResultRole='GroupComponent',
                                  ResultClass='LMI_UnixDirectory')
            self.assertEquals(assocs, [])

            ## GroupComponent - LMI_UnixDirectory
            if isinstance(system, pywbem.CIMInstanceName):
                system.path = system
            assocs = assoc_method(system.path,
                                  AssocClass=assoc_class,
                                  Role='GroupComponent',
                                  ResultRole='PartComponent',
                                  ResultClass='LMI_UnixDirectory')
            self.assertEquals(len(assocs), 1)
            self.assertEquals(assocs[0]['Name'], '/')
            # wrong Role
            assocs = assoc_method(system.path, AssocClass=assoc_class, Role='PartComponent')
            self.assertEquals(assocs, [])
            # wrong ResultRole
            assocs = assoc_method(system.path, AssocClass=assoc_class, ResultRole='GroupComponent')
            self.assertEquals(assocs, [])
            # good role and ResultRole, wrong ResultClass
            assocs = assoc_method(system.path,
                                  AssocClass=assoc_class,
                                  Role='GroupComponent',
                                  ResultRole='PartComponent',
                                  ResultClass=self.system_cs_name)
            self.assertEquals(assocs, [])


        ### References and ReferenceNames
        for assoc_method in [self.wbemconnection.References, self.wbemconnection.ReferenceNames]:
            ## PartComponent - CIM_ComputerSystem
            assocs = assoc_method(cop, ResultClass=assoc_class)
            self.assertEquals(len(assocs), 1)
            system = assocs[0]['GroupComponent']
            self.assertEquals(system['CreationClassName'], self.system_cs_name)
            self.assertEquals(system['Name'], self.SYSTEM_NAME)

            ## GroupComponent - LMI_UnixDirectory
            assocs = assoc_method(system, ResultClass=assoc_class)
            self.assertEquals(len(assocs), 1)
            self.assertEquals(assocs[0]['PartComponent']['Name'], '/')

        ### EnumerateInstances and GetInstance
        insts = self.wbemconnection.EnumerateInstances(assoc_class)
        self.assertEquals(len(insts), 1)
        system = insts[0]['GroupComponent']
        rootdir = insts[0]['PartComponent']
        self.assertEquals(system['CreationClassName'], self.system_cs_name)
        self.assertEquals(system['Name'], self.SYSTEM_NAME)
        self.assertEquals(rootdir['CSCreationClassName'], self.system_cs_name)
        self.assertEquals(rootdir['CSName'], self.SYSTEM_NAME)
        self.assertEquals(rootdir['Name'], '/')

        inst_cop = pywbem.CIMInstanceName(classname=assoc_class,
                                          namespace='root/cimv2',
                                          keybindings={
                                              'GroupComponent':system,
                                              'PartComponent':rootdir
                                          })
        inst = self.wbemconnection.GetInstance(inst_cop)
        self.assertTrue(inst is not None)
        system = inst['GroupComponent']
        rootdir = inst['PartComponent']
        self.assertEquals(system['CreationClassName'], self.system_cs_name)
        self.assertEquals(system['Name'], self.SYSTEM_NAME)
        self.assertEquals(rootdir['CSCreationClassName'], self.system_cs_name)
        self.assertEquals(rootdir['CSName'], self.SYSTEM_NAME)
        self.assertEquals(rootdir['Name'], '/')

    def test_mkdir(self):
        def mkdir(path):
            cop = self.cop.copy()
            cop['Name'] = path
            inst = pywbem.CIMInstance('LMI_UnixDirectory', cop.keybindings)
            self.wbemconnection.CreateInstance(inst)

        try:
            mkdir(self.testdir + '/mkdir-test')
        except pywbem.CIMError as pe:
            self.fail(pe[1])

        self.assertRaises(pywbem.CIMError,
                          mkdir,
                          '/cant/create/me')

    def test_rmdir(self):
        def rmdir(path):
            cop = self.cop.copy()
            cop['Name'] = path
            inst = pywbem.CIMInstance('LMI_UnixDirectory', cop.keybindings)
            self.wbemconnection.DeleteInstance(cop)

        try:
            rmdir(self.files['dir']['path'])
        except pywbem.CIMError as pe:
            self.fail(pe[1])

        self.assertRaises(pywbem.CIMError,
                          rmdir,
                          '/cant/remove/me')

    # for now, this test just checks if FSName, FSCreationClassName,
    # CreationClassName and LFCreationClassName are properly ignored; empty
    # strings are used, since they should represent "ignored" pretty well
    def _test_missing_or_wrong_properties(self, is_unixfile):
        testfile = self.files['data']
        if is_unixfile:
            prefix = 'LF'
            clsname = 'LMI_UnixFile'
        else:
            prefix = ''
            clsname = 'LMI_DataFile'
        cop = pywbem.CIMInstanceName(classname=clsname,
                                     namespace='root/cimv2',
                                     keybindings={
                                         'CSCreationClassName':self.system_cs_name,
                                         'CSName':self.SYSTEM_NAME,
                                         'FSCreationClassName':'',
                                         'FSName':'',
                                     })

        cop.keybindings[prefix+'Name'] = testfile['path']
        cop.keybindings[prefix+'CreationClassName'] = ''

        try:
            self.wbemconnection.GetInstance(cop)
        except pywbem.CIMError as pe:
            self.fail(pe[1])

    def test_unixfile_missing_or_wrong_properties(self):
        self._test_missing_or_wrong_properties(True)

    def test_logicalfile_missing_or_wrong_properties(self):
        self._test_missing_or_wrong_properties(False)

    def test_transient_file(self):
        cop = self.cop.copy()
        cop['Name'] = self.transient_file['path']

        try:
            inst = self.wbemconnection.GetInstance(cop)
        except pywbem.CIMError as pe:
            self.fail(pe[1])

        self.assertEquals(inst['FSCreationClassName'], 'LMI_TransientFileSystem')
        self.assertEquals(inst['FSName'], 'PATH=' + cop['Name'])

if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
#
# Copyright (C) 2012-2013 Red Hat, Inc.  All rights reserved.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#
# Authors: Michal Minar <miminar@redhat.com>
# Authors: Jan Grec <jgrec@redhat.com>
#
"""
Unit tests for ``LMI_SoftwareInstallationService`` provider.
"""

import pywbem
import time
import unittest

from lmi.test.lmibase import enable_lmi_exceptions
import package
import swbase
import util

CHECK_SOFTWARE_IDENTITY_NOT_SUPPORTED = 1
COMMUNICATION_STATUS_NOT_AVAILABLE = 1
COMMUNICATION_STATUS_OK = 2
DETAILED_STATUS_NOT_AVAILABLE = 0
ENABLED_DEFAULT_ENABLED = 2
ENABLED_DEFAULT_NOT_APPLICABLE = 5
ENABLED_STATE_ENABLED = 2
ENABLED_STATE_NOT_APPLICABLE = 5
FIND_IDENTITY_FOUND = 0
FIND_IDENTITY_NO_MATCH = 1
HEALTH_STATE_OK = 5
INSTALL_FROM_BYTE_STREAM_NOT_SUPPORTED = 1
INSTALL_OPTIONS_FORCE = 3
INSTALL_OPTIONS_INSTALL = 4
INSTALL_OPTIONS_UPDATE = 5
INSTALL_OPTIONS_REPAIR = 6
INSTALL_OPTIONS_UNINSTALL = 9
OPERATING_STATUS_SERVICING = 2
OPERATIONAL_STATUS_OK = 2
PRIMARY_STATUS_OK = 1
REQUESTED_STATE_NOT_APPLICABLE = 12
TRANSITIONING_TO_STATE_NOT_APPLICABLE = 12

JOB_COMPLETED_WITH_NO_ERROR = 0
UNSPECIFIED_ERROR = 2
INVALID_PARAMETER = 5
METHOD_PARAMETERS_CHECKED_JOB_STARTED = 4096

class TestSoftwareInstallationService(swbase.SwTestCase):
    """
    Basic cim operations test on ``LMI_SoftwareInstallationService``.
    """

    CLASS_NAME = "LMI_SoftwareInstallationService"
    KEYS = ("CreationClassName", "Name", "SystemCreationClassName",
            "SystemName")

    def make_op(self):
        """
        :returns: Object path of ``LMI_SoftwareInstallationService``
        :rtype: :py:class:`lmi.shell.LMIInstanceName`
        """
        return self.cim_class.new_instance_name({
            "Name" : 'LMI:' + self.CLASS_NAME,
            "SystemName" : self.SYSTEM_NAME,
            "SystemCreationClassName" : self.system_cs_name,
            "CreationClassName" : self.CLASS_NAME
        })

    def test_get_instance(self):
        """
        Test ``GetInstance()`` call on ``LMI_SoftwareInstallationService``.
        """
        objpath = self.make_op()
        inst = objpath.to_instance()
        self.assertNotEqual(inst, None)
        self.assertCIMNameEqual(inst.path, objpath)
        for key_property in objpath.key_properties():
            if key_property == 'SystemName':
                self.assertIn(inst.SystemName, (self.SYSTEM_NAME, 'localhost'))
            else:
                self.assertEqual(getattr(inst, key_property),
                        getattr(objpath, key_property))
        self.assertGreater(len(inst.Caption), 0)
        self.assertEqual(inst.CommunicationStatus,
                COMMUNICATION_STATUS_NOT_AVAILABLE)
        self.assertGreater(len(inst.Description), 0)
        self.assertEqual(inst.DetailedStatus, DETAILED_STATUS_NOT_AVAILABLE)
        self.assertEqual(inst.EnabledDefault, ENABLED_DEFAULT_NOT_APPLICABLE)
        self.assertEqual(inst.EnabledState, ENABLED_STATE_NOT_APPLICABLE)
        self.assertEqual(inst.HealthState, HEALTH_STATE_OK)
        self.assertEqual(inst.InstanceID, "LMI:" + self.CLASS_NAME)
        self.assertEqual(inst.OperatingStatus, OPERATING_STATUS_SERVICING)
        self.assertEqual(inst.OperationalStatus, [OPERATIONAL_STATUS_OK])
        self.assertEqual(inst.PrimaryStatus, PRIMARY_STATUS_OK)
        self.assertEqual(inst.RequestedState, REQUESTED_STATE_NOT_APPLICABLE)
        self.assertTrue(inst.Started)
        self.assertEqual(inst.TransitioningToState, 12)

    def test_enum_instance_names(self):
        """
        Test ``EnumerateInstanceNames()`` call on
        ``LMI_SoftwareInstallationService``.
        """
        objpath = self.make_op()
        insts = self.cim_class.instance_names()
        self.assertEqual(len(insts), 1)
        inst = insts[0]
        self.assertCIMNameEqual(inst, objpath)

    def test_enum_instances(self):
        """
        Test ``EnumerateInstances()`` call on
        ``LMI_SoftwareInstallationService``.
        """
        objpath = self.make_op()
        insts = self.cim_class.instances()
        self.assertEqual(len(insts), 1)
        inst = insts[0]
        self.assertCIMNameEqual(inst.path, objpath)
        self.assertEqual(inst.InstanceID, "LMI:" + self.CLASS_NAME)

    @enable_lmi_exceptions
    def test_check_software_identity_method(self):
        """
        Try to invoke ``CheckSoftwareIdentity()`` method which shall not be
        supported.
        """
        service = self.make_op().to_instance()
        # TODO: This should either return NOT_SUPPORTED or
        # InstallFromByteStream() should raise the same.
        # This dissimilarity is really weird.
        self.assertRaisesCIM(pywbem.CIM_ERR_METHOD_NOT_AVAILABLE,
                service.CheckSoftwareIdentity)

    @enable_lmi_exceptions
    def test_install_from_byte_stream_method(self):
        """
        Try to invoke ``InstallFromByteStram()`` method which shall not be
        supported.
        """
        service = self.make_op().to_instance()
        # TODO: This should either raise METHOD_NOT_AVAILABLE or
        # CheckSoftwareIdentity() should return NOT_SUPPORTED.
        # This dissimilarity is really weird.
        rval, _, _ = service.InstallFromByteStream()
        self.assertEqual(rval, INSTALL_FROM_BYTE_STREAM_NOT_SUPPORTED)

    @swbase.test_with_repos(**{
        'stable' : True,
        'updates-testing' : True,
        'updates' : False,
        'misc' : False,
    })
    @swbase.test_with_packages(**{
        'stable#pkg1' : True,
        'pkg2' : False,
        'pkg3' : False,
        'pkg4' : False,
        'misc#*' : False,
    })
    def test_query_identities_by_name(self):
        """
        Try to find packages by their name.
        """
        service = self.make_op().to_instance()

        # allow duplicates is False
        rval, oparms, _ = service.FindIdentity(Name="openlmi-sw-test")
        self.assertEqual(rval, FIND_IDENTITY_FOUND)
        self.assertIn("Matches", oparms)
        inames = oparms["Matches"]
        # pkg1 is listed twice (installed and available version)
        # pkg2, pkg3 and pkg4 just once (the newest version)
        self.assertEqual(len(inames), 5)
        nevra_set = set()
        for iname in inames:
            self.assertEqual(iname.classname, 'LMI_SoftwareIdentity')
            self.assertEqual(iname.key_properties(), ['InstanceID'])
            self.assertTrue(iname.InstanceID.startswith(
                'LMI:LMI_SoftwareIdentity'))
            nevra_set.add(iname.InstanceID[len('LMI:LMI_SoftwareIdentity:'):])
        for repoid in ('stable', 'updates-testing'):
            for pkg in self.get_repo(repoid).packages:
                if pkg.name.endswith('pkg2') and repoid == 'stable':
                    # only pkg2 from updates-tesing will be present (its newer)
                    continue
                self.assertIn(pkg.nevra, nevra_set)
                nevra_set.remove(pkg.nevra)
        self.assertEqual(len(nevra_set), 0)

        rval, oparms, _ = service.FindIdentity(Name="openlmi-sw-test",
                AllowDuplicates=True)
        self.assertEqual(rval, FIND_IDENTITY_FOUND)
        inames = oparms["Matches"]
        self.assertEqual(len(inames), 6)
        nevra_set = set()
        for iname in inames:
            nevra_set.add(iname.InstanceID[len('LMI:LMI_SoftwareIdentity:'):])
        for repoid in ('stable', 'updates-testing'):
            for pkg in self.get_repo(repoid).packages:
                self.assertIn(pkg.nevra, nevra_set)
                nevra_set.remove(pkg.nevra)
        self.assertEqual(len(nevra_set), 0)

    @swbase.test_with_repos('stable', 'updates', 'updates-testing', 'misc')
    def test_query_not_existing_package_by_name(self):
        """
        Try to find package by part of name with exact match.
        """
        service = self.make_op().to_instance()
        rval, oparms, _ = service.FindIdentity(Name="openlmi-sw-test",
                ExactMatch=True)
        self.assertEqual(rval, FIND_IDENTITY_NO_MATCH)
        self.assertIn("Matches", oparms)
        inames = oparms["Matches"]
        self.assertEqual(len(inames), 0)

    @swbase.test_with_repos('stable', 'updates', 'updates-testing', 'misc')
    @swbase.test_with_packages(**{'pkg2' : False})
    def test_query_identity_by_name_exact(self):
        """
        Try to find package by name with exact match.
        """
        service = self.make_op().to_instance()

        # allow duplicates is False
        rval, oparms, _ = service.FindIdentity(Name="openlmi-sw-test-pkg2",
                ExactMatch=True)
        self.assertEqual(rval, FIND_IDENTITY_FOUND)
        self.assertIn("Matches", oparms)
        inames = oparms["Matches"]
        # there are two different architectures available
        self.assertEqual(len(inames), 2)
        nevra_set = set(   i.InstanceID[len('LMI:LMI_SoftwareIdentity:'):]
                       for i in inames)
        self.assertIn(self.get_repo('updates-testing')['pkg2'].nevra, nevra_set)
        self.assertIn(self.get_repo('misc')['pkg2'].nevra, nevra_set)

        # allow duplicates is True
        rval, oparms, _ = service.FindIdentity(Name="openlmi-sw-test-pkg2",
                AllowDuplicates=True,
                ExactMatch=True)
        self.assertEqual(rval, FIND_IDENTITY_FOUND)
        self.assertIn("Matches", oparms)
        inames = oparms["Matches"]
        self.assertEqual(len(inames), 4)
        nevra_set = set(   i.InstanceID[len('LMI:LMI_SoftwareIdentity:'):]
                       for i in inames)
        for repo in self.repodb.values():
            pkg = repo['pkg2']
            self.assertIn(pkg.nevra, nevra_set)
            nevra_set.remove(pkg.nevra)
        self.assertEqual(len(nevra_set), 0)

    @swbase.test_with_repos('stable', 'updates', 'updates-testing', 'misc')
    def test_query_identity_by_repoid(self):
        """
        Try to filter packages by repository.
        """
        service = self.make_op().to_instance()
        repo = self.get_repo('misc')
        repo_iname = self.ns.LMI_SoftwareIdentityResource.new_instance_name({
            "Name" : repo.repoid,
            "SystemName" : self.SYSTEM_NAME,
            "SystemCreationClassName" : self.system_cs_name,
            "CreationClassName" : "LMI_SoftwareIdentityResource"
        })
        rval, oparms, _ = service.FindIdentity(Repository=repo_iname)
        self.assertEqual(rval, FIND_IDENTITY_FOUND)
        self.assertIn("Matches", oparms)
        inames = oparms['Matches']
        self.assertEqual(len(inames), len(repo.packages))
        nevra_set = set(   i.InstanceID[len('LMI:LMI_SoftwareIdentity:'):]
                       for i in inames)
        for pkg in repo.packages:
            self.assertIn(pkg.nevra, nevra_set)
            nevra_set.remove(pkg.nevra)
        self.assertEqual(len(nevra_set), 0)

    @swbase.test_with_repos(**{'stable' : False, 'misc' : False})
    @swbase.test_with_packages('stable#pkg1')
    def test_query_installed_package_by_nevra(self):
        """
        Try to find installed, not available package by its nevra.
        """
        service = self.make_op().to_instance()
        pkg = self.get_repo('stable')['pkg1']
        rval, oparms, _ = service.FindIdentity(
                Name=pkg.name,
                Epoch=pkg.epoch,
                Version=pkg.ver,
                Release=pkg.rel,
                Architecture=pkg.arch,
                AllowDuplicates=True,
                ExactMatch=True)
        self.assertEqual(rval, FIND_IDENTITY_FOUND)
        self.assertIn("Matches", oparms)
        inames = oparms['Matches']
        self.assertEqual(len(inames), 1)
        self.assertTrue(inames[0].InstanceID.endswith(pkg.nevra))

    @swbase.test_with_repos('stable', 'updates', 'updates-testing', 'misc')
    def test_query_package_by_arch(self):
        """
        Try to find package filtered by arch.
        """
        service = self.make_op().to_instance()
        pkg = self.get_repo('misc')['pkg2']
        rval, oparms, _ = service.FindIdentity(
                Name='openlmi-sw-test',
                Architecture=pkg.arch,
                AllowDuplicates=True)
        self.assertEqual(rval, FIND_IDENTITY_FOUND)
        self.assertIn("Matches", oparms)
        inames = oparms['Matches']
        self.assertEqual(len(inames), 1)
        self.assertTrue(inames[0].InstanceID.endswith(pkg.nevra))

    @swbase.test_with_repos('stable')
    @swbase.test_with_packages(**{ 'pkg1' : False })
    def test_install_package_sync(self):
        """
        Try to synchronously install package.
        """
        pkg = self.get_repo('stable')['pkg1']
        self.assertFalse(package.is_pkg_installed(pkg))
        service = self.make_op().to_instance()
        rval, oparms, _ = service.SyncInstallFromSoftwareIdentity(
                Source=util.make_pkg_op(self.ns, pkg),
                InstallOptions=[INSTALL_OPTIONS_INSTALL],
                Target=self.system_iname)
        self.assertEqual(rval, JOB_COMPLETED_WITH_NO_ERROR)
        self.assertEqual(len(oparms), 0)
        self.assertTrue(package.is_pkg_installed(pkg))

    @swbase.test_with_repos('stable')
    @swbase.test_with_packages('stable#pkg1')
    def test_remove_package_sync(self):
        """
        Try to synchronously remove package.
        """
        pkg = self.get_repo('stable')['pkg1']
        self.assertTrue(package.is_pkg_installed(pkg))
        service = self.make_op().to_instance()
        rval, oparms, _ = service.SyncInstallFromSoftwareIdentity(
                Source=util.make_pkg_op(self.ns, pkg),
                InstallOptions=[INSTALL_OPTIONS_UNINSTALL],
                Target=self.system_iname)
        self.assertEqual(rval, JOB_COMPLETED_WITH_NO_ERROR)
        self.assertTrue(  len(oparms) == 0
                       or ('Job' in oparms and oparms['Job'] is None))
        self.assertFalse(package.is_pkg_installed(pkg))

    @swbase.test_with_repos('stable', 'updates')
    @swbase.test_with_packages('stable#pkg1')
    def test_update_package_sync(self):
        """
        Try to synchronously update package.
        """
        pkg = self.get_repo('stable')['pkg1']
        self.assertTrue(package.is_pkg_installed(pkg))
        service = self.make_op().to_instance()
        rval, oparms, _ = service.SyncInstallFromSoftwareIdentity(
                Source=util.make_pkg_op(self.ns, pkg),
                InstallOptions=[INSTALL_OPTIONS_UPDATE],
                Target=self.system_iname)
        self.assertEqual(rval, JOB_COMPLETED_WITH_NO_ERROR)
        self.assertTrue(  len(oparms) == 0
                       or ('Job' in oparms and oparms['Job'] is None))
        self.assertFalse(package.is_pkg_installed(pkg))
        up_pkg = self.get_repo('updates')['pkg1']
        self.assertTrue(package.is_pkg_installed(up_pkg))

    @swbase.test_with_repos('stable', 'updates')
    @swbase.test_with_packages(**{'pkg1' : False})
    def test_update_not_installed_package_sync(self):
        """
        Try to synchronously update package which is not installed.
        """
        pkg = self.get_repo('stable')['pkg1']
        self.assertFalse(package.is_pkg_installed(pkg.name))
        service = self.make_op().to_instance()
        rval, oparms, error = service.SyncInstallFromSoftwareIdentity(
                Source=util.make_pkg_op(self.ns, pkg),
                InstallOptions=[INSTALL_OPTIONS_UPDATE],
                Target=self.system_iname)
        self.assertEqual(rval, INVALID_PARAMETER)
        self.assertGreater(len(error), 0)
        self.assertTrue(  len(oparms) == 0
                       or ('Job' in oparms and oparms['Job'] is None))
        self.assertFalse(package.is_pkg_installed(pkg.name))

    @enable_lmi_exceptions
    @swbase.test_with_repos('stable')
    @swbase.test_with_packages(**{ 'pkg1' : False })
    def test_install_method_sync_without_target_and_collection(self):
        """
        Try to synchronously install package without target and collection
        parameters given.

        Software Update profile says that ``InstallFromSoftwareIdentity``
        shall return ``UNSPECIFIED_ERROR``. Which is really weird since the
        error is known.
        """
        pkg = self.get_repo('stable')['pkg1']
        self.assertFalse(package.is_pkg_installed(pkg))
        service = self.make_op().to_instance()
        for opts in (
                [INSTALL_OPTIONS_INSTALL],
                [INSTALL_OPTIONS_UNINSTALL],
                [INSTALL_OPTIONS_UPDATE]):
            rval, oparms, _ = service.SyncInstallFromSoftwareIdentity(
                    Source=util.make_pkg_op(self.ns, pkg),
                    InstallOptions=opts)
            self.assertEqual(rval, UNSPECIFIED_ERROR)
            self.assertTrue(  len(oparms) == 0
                           or ('Job' in oparms and oparms['Job'] is None))

    @swbase.test_with_repos('stable')
    @swbase.test_with_packages(**{ 'pkg1' : False })
    def test_install_package_sync_with_target_and_collection(self):
        """
        Try to synchronously install package with target and collection
        parameters omitted.

        Software Update profile says that ``InstallFromSoftwareIdentity``
        shall return ``UNSPECIFIED_ERROR``. Which is really weird since the
        error is known.
        """
        pkg = self.get_repo('stable')['pkg1']
        self.assertFalse(package.is_pkg_installed(pkg))
        service = self.make_op().to_instance()
        for opts in (
                [INSTALL_OPTIONS_INSTALL],
                [INSTALL_OPTIONS_UNINSTALL],
                [INSTALL_OPTIONS_UPDATE]):
            rval, oparms, _ = service.SyncInstallFromSoftwareIdentity(
                    Source=util.make_pkg_op(self.ns, pkg),
                    InstallOptions=opts,
                    Target=self.system_iname,
                    Collection=self.ns.LMI_SystemSoftwareCollection \
                        .new_instance_name({
                            "InstanceID" : "LMI:LMI_SystemSoftwareCollection"
                        }))
            self.assertEqual(rval, UNSPECIFIED_ERROR)
            self.assertTrue(  len(oparms) == 0
                           or ('Job' in oparms and oparms['Job'] is None))

    @swbase.test_with_repos('stable')
    @swbase.test_with_packages(**{ 'pkg1' : False })
    def test_install_package_sync_with_collection(self):
        """
        Try to synchronously install package with just collection given.
        """
        pkg = self.get_repo('stable')['pkg1']
        self.assertFalse(package.is_pkg_installed(pkg))
        service = self.make_op().to_instance()
        rval, oparms, _ = service.SyncInstallFromSoftwareIdentity(
                Source=util.make_pkg_op(self.ns, pkg),
                InstallOptions=[INSTALL_OPTIONS_INSTALL],
                Collection=self.ns.LMI_SystemSoftwareCollection \
                        .new_instance_name({
                            "InstanceID" : "LMI:LMI_SystemSoftwareCollection"
                        }))
        self.assertEqual(rval, JOB_COMPLETED_WITH_NO_ERROR)
        self.assertTrue(  len(oparms) == 0
                       or ('Job' in oparms and oparms['Job'] is None))
        self.assertTrue(package.is_pkg_installed(pkg))

    @swbase.test_with_repos('stable')
    @swbase.test_with_packages('stable#pkg1')
    def test_reinstall_package(self):
        """
        Try to reinstall package.
        """
        pkg = self.get_repo('stable')['pkg1']
        pkg_op = util.make_pkg_op(self.ns, pkg)
        inst = pkg_op.to_instance()
        install_date = inst.InstallDate
        self.assertTrue(package.is_pkg_installed(pkg))

        service = self.make_op().to_instance()
        time.sleep(1)
        rval, oparms, _ = service.SyncInstallFromSoftwareIdentity(
                Source=pkg_op,
                InstallOptions=[INSTALL_OPTIONS_INSTALL, INSTALL_OPTIONS_FORCE],
                Target=self.system_iname)
        self.assertEqual(rval, JOB_COMPLETED_WITH_NO_ERROR)
        self.assertTrue(  len(oparms) == 0
                       or ('Job' in oparms and oparms['Job'] is None))
        self.assertTrue(package.is_pkg_installed(pkg))
        inst.refresh()
        install_date2 = inst.InstallDate
        self.assertGreater(install_date2, install_date)

def suite():
    """For unittest loaders."""
    return unittest.TestLoader().loadTestsFromTestCase(
            TestSoftwareInstallationService)

if __name__ == '__main__':
    unittest.main()



# Software Management Providers
#
# Copyright (C) 2012-2013 Red Hat, Inc.  All rights reserved.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#
# Authors: Michal Minar <miminar@redhat.com>
#

"""
Common utilities meant to be used only be ``yumdb`` subpackage.
"""

import logging
import yum

from lmi.providers import cmpi_logging
from lmi.software.util import Configuration

DEBUG_LOGGING_CONFIG = {
    "version" : 1,
    'disable_existing_loggers' : True,
    "formatters": {
        # this is a message format for logging function/method calls
        # it's manually set up in YumWorker's init method
        "default": {
            "()": "lmi.providers.cmpi_logging.DispatchingFormatter",
            "formatters" : {
                "lmi.providers.cmpi_logging.trace_function_or_method":
                    "%(asctime)s %(levelname)s:%(message)s"
                },
            "default" : "%(asctime)s %(levelname)s:%(module)s:"
                        "%(funcName)s:%(lineno)d - %(message)s"
            },
        },
    "handlers": {
        "file" : {
            "class" : "logging.FileHandler",
            "filename" : "/var/tmp/YumWorker.log",
            "level" : "DEBUG",
            "formatter": "default",
            },
        },
    "root": {
        "level": "DEBUG",
        "handlers" : ["file"]
        },
}

DISABLED_LOGGING_CONFIG = {
    'version' : 1,
    'disable_existing_loggers' : True,
    'handlers': {
        'null' : {
            'class': 'logging.FileHandler',
            'level': 'CRITICAL',
            'filename': '/dev/null'
            }
        },
    'root' : {
        'level': 'CRITICAL',
        'handlers' : ['null'],
        }
}

def setup_logging():
    """
    This is meant to be used by ``YumWorker`` process to setup logging
    independent of what providers are using. Unfortunately ``YumWorker``
    can not use the same facilities as the rest of program, because
    logging is done through *broker*.
    """
    config = Configuration.get_instance()
    cp = config.config
    logging_setup = False
    if cp.has_option('YumWorkerLog', 'file_config'):
        try:
            file_path = config.file_path('YumWorkerLog', 'FileConfig')
            logging.config.fileConfig(file_path)
            logging_setup = True
        except Exception:
            pass
    if logging_setup is False:
        out = config.get_safe('YumWorkerLog', 'OutputFile')
        if out is not None:
            try:
                defaults = DEBUG_LOGGING_CONFIG.copy()
                defaults["handlers"]["file"]["filename"] = out
                level = config.get_safe('YumWorkerLog', 'Level')
                level = cmpi_logging.LOGGING_LEVELS[level.lower()]
                defaults["handlers"]["file"]["level"] = level
                defaults["root"]["level"] = level
                logging.config.dictConfig(defaults)
                logging_setup = True
                # re-enable function tracing logger which has been disabled by
                # dictConfig() call
                logging.getLogger(cmpi_logging.__name__
                        + '.trace_function_or_method').disabled = False
            except Exception:
                pass
    if logging_setup is False:
        # disable logging completely
        logging.config.dictConfig(DISABLED_LOGGING_CONFIG)

def is_pkg_installed(pkg, rpmdb):
    """
    Returns ``True`` if the package is installed.

    :param pkg: Yum package object.
    :type pkg: :py:class:`yum.packages.PackageObject`
    :param rpmdb: Installed package sack.
    :type rpmdb: :py:class:`yum.rpmsack.RPMDBPackageSack`
    :returns: Whether the package is installed or not.
    :rtype: boolean
    """
    if not isinstance(pkg, yum.packages.PackageObject):
        raise TypeError("pkg must be a PackageObject")
    return (  isinstance(pkg, yum.rpmsack.RPMInstalledPackage)
           or pkg.pkgtup in rpmdb._tup2pkg)



# -*- encoding: utf-8 -*-
# Software Management Providers
#
# Copyright (C) 2012-2013 Red Hat, Inc.  All rights reserved.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA

"""
Just a common functionality related to class LMI_SoftwareMethodResult.
"""

import pywbem

from lmi.providers import cmpi_logging
from lmi.software import util
from lmi.software.core import InstMethodCall
from lmi.software.core import Job
from lmi.software.yumdb import jobs, errors, YumDB

@cmpi_logging.trace_function
def object_path2jobid(op):
    """
    @param op must contain precise InstanceID of job
    """
    if not isinstance(op, pywbem.CIMInstanceName):
        raise TypeError("op must be a CIMInstanceName")
    if not op["InstanceID"].lower().startswith('lmi:lmi_softwaremethodresult:'):
        raise pywbem.CIMError(pywbem.CIM_ERR_NOT_FOUND,
                "Missing 'LMI:LMI_SoftwareMethodResult:' prefix in InstanceID.")
    instid = op['InstanceID'][len('LMI:LMI_SoftwareMethodResult:'):]
    try:
        instid = int(instid)
    except ValueError:
        raise pywbem.CIMError(pywbem.CIM_ERR_NOT_FOUND,
                "Failed to parse InstanceID.")
    return instid

@cmpi_logging.trace_function
def object_path2job(op):
    """
    @param op must contain precise InstanceID of job
    """
    instid = object_path2jobid(op)
    try:
        return YumDB.get_instance().get_job(instid)
    except errors.JobNotFound:
        raise pywbem.CIMError(pywbem.CIM_ERR_NOT_FOUND, "No such job.")

@cmpi_logging.trace_function
def job2model(env, job, keys_only=True, model=None):
    """
    This accepts job's id and produces corresponding result model.
    """
    if not isinstance(job, jobs.YumJob):
        raise TypeError("job must be an instance of YumJob")
    if model is None:
        model = util.new_instance_name("LMI_SoftwareMethodResult")
        if not keys_only:
            model = pywbem.CIMInstance("LMI_SoftwareMethodResult", path=model)
    model['InstanceID'] = "LMI:LMI_SoftwareMethodResult:"+str(job.jobid)
    if not keys_only:
        model.path['InstanceID'] = model['InstanceID']  #pylint: disable=E1103
        method_name = Job.JOB_METHOD_NAMES[job.metadata['method']]
        model['Caption'] = 'Result of method %s' % method_name
        model['Description'] = (
                'Result of asynchronous job number %d created upon invocation'
                " of %s's %s method." % (job.jobid,
                "LMI_SoftwareInstallationService", method_name))
        model['ElementName'] = 'MethodResult-%d' % job.jobid
        model['PostCallIndication'] = pywbem.CIMProperty("PostCallIndication",
                type="instance",
                value=InstMethodCall.job2model(env, job, pre=False))
        model['PreCallIndication'] = pywbem.CIMProperty("PreCallIndication",
                type="instance",
                value=InstMethodCall.job2model(env, job))
    return model




# Software Management Providers
#
# Copyright (C) 2012-2013 Red Hat, Inc.  All rights reserved.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#
# Authors: Michal Minar <miminar@redhat.com>
#
"""
This is a module for ``JobManager`` which is a standalone thread in provider
process. It keeps a cache of asynchronous jobs and handles input queue of jobs
that are about to be handled by
:py:class:`~lmi.software.yumdb.process.YumWorker` process.

Before using ``JobManager``, module's variable ``JOB_TO_MODEL`` should be set
to a callable taking :py:class:`~lmi.software.yumdb.jobs.YumJob` instance and
returning its matching CIM abstraction instance.
"""
from functools import wraps
import heapq
import inspect
import sys
import threading
import time
import traceback

from lmi.providers import cmpi_logging
from lmi.providers.IndicationManager import IndicationManager
from lmi.providers.JobManager import JobManager as JM
from lmi.software.yumdb import errors
from lmi.software.yumdb import jobs
from lmi.software.yumdb.sessionmanager import SessionManager
from lmi.software.util import Configuration

# This is a callable, which must be initialized before JobManager is used.
# It should be a pointer to function, which takes a job and returns
# corresponding CIM instance. It's used for sending indications.
JOB_TO_MODEL = lambda job: None

LOG = cmpi_logging.get_logger(__name__)

# *****************************************************************************
# Decorators
# *****************************************************************************
def job_handler(job_from_target=True):
    """
    Decorator for JobManager methods serving as handlers for control jobs.

    Decorator locks the :py:attr:`JobManager._job_lock` of manager's instance.
    """
    def _wrapper_jft(method):
        """
        It consumes "target" keyword argument (which is job's id) and makes
        it an instance of YumJob. The method is then called with "job" argument
        instead of "target".
        """
        logged = cmpi_logging.trace_method(method, frame_level=2)

        @wraps(method)
        def _new_func(self, *args, **kwargs):
            """Wrapper around method."""
            if 'target' in kwargs:
                kwargs['job'] = kwargs.pop('target')
            callargs = inspect.getcallargs(method, self, *args, **kwargs)
            target = callargs.pop('job')
            with self._job_lock:
                if not target in self._async_jobs:
                    raise errors.JobNotFound(target)
                job = self._async_jobs[target]
                callargs['job'] = job
                return logged(**callargs)
        return _new_func

    def _simple_wrapper(method):
        """Just locks the job lock."""
        @wraps(method)
        def _new_func(self, *args, **kwargs):
            """Wrapper around method."""
            with self._job_lock:
                return method(self, *args, **kwargs)
        return _new_func

    if job_from_target:
        return _wrapper_jft
    else:
        return _simple_wrapper

class JobIndicationSender(object):
    """
    Makes creation and sending of indications easy. It keeps a reference
    to job, which can be *snapshotted* for making CIM instance out of it.
    These instances are then used to send indications via
    :py:class:`lmi.providers.IndicationManager.IndicationManager`.

    Typical usage::

        sender = JobIndicationSender(im, job, [fltr_id1, fltr_id2])
            ... # modify job
        sender.snapshot()
        sender.send()

    .. note::
        Number of kept CIM instances won't exceed 2. First one is created upon
        instantiation and the second one with the subsequent call to
        :py:meth:`snapshot()`. Any successive call to it will overwrite the
        second instance.

    :param job: Is job instance, which will be immediately snapshoted as old
        instance and later as a new one.
    :type job: :py:class:`~lmi.software.yumdb.jobs.YumJob`
    :param list indications: Can either be a list of indication ids or a single
        indication id.
    :param new: A job instance stored as new.
    :type new: :py:class:`~lmi.software.yumdb.jobs.YumJob`
    """

    def __init__(self, indication_manager, job,
            indications=JM.IND_JOB_CHANGED, new=None):
        if not isinstance(indication_manager, IndicationManager):
            raise TypeError("indication_manager must be a subclass of"
                    " IndicationManager")
        if not isinstance(job, jobs.YumJob):
            raise TypeError("job must be an instance of YumJob")
        if not new is None and not isinstance(new, jobs.YumJob):
            raise TypeError("new must be an instance of YumJob")
        self._indication_manager = indication_manager
        self._job = job
        self._old_instance = JOB_TO_MODEL(job)
        if new is not None:
            new = JOB_TO_MODEL(job)
        self._new_instance = new
        self._indications = set()
        self.indication_ids = indications

    @property
    def job(self):
        """
        :returns: Associated job object.
        :rtype: :py:class:`~lmi.software.yumdb.jobs.YumJob`
        """
        return self._job

    @property
    def indication_ids(self):
        """
        :returns: Set of indication filter IDs.
        :rtype: set
        """
        return self._indications.copy()

    @indication_ids.setter
    def indication_ids(self, indication_ids):
        """
        Set the indication filter IDs.

        :param list indication_ids: Can be even single id.
        """
        if isinstance(indication_ids, basestring):
            indication_ids = set([indication_ids])
        self._indications = set(indication_ids)

    @cmpi_logging.trace_method
    def add_indication_ids(self, indication_ids):
        """
        Add filter IDs.
        """
        if isinstance(indication_ids, basestring):
            indication_ids = set([indication_ids])
        self._indications.update(indication_ids)

    @cmpi_logging.trace_method
    def snapshot(self):
        """
        Make a second CIM instance, overwriting previous one (not the first).
        """
        self._new_instance = JOB_TO_MODEL(self._job)

    @cmpi_logging.trace_method
    def send(self, make_snapshot=False):
        """
        Send all requested indications for given job.
        """
        if not self._indications:
            raise errors.IndicationError(
                    "can not send any indication without id")
        if make_snapshot:
            self.snapshot()
        if (   JM.IND_JOB_CHANGED in self._indications
           and self._new_instance is None):
            raise errors.IndicationError("no snapshot made for modified job")
        indication_manager = IndicationManager.get_instance()
        for fltr_id in self._indications:
            if fltr_id == JM.IND_JOB_CREATED:
                LOG().debug("sending instance creation indication for job %s",
                        self._job)
                indication_manager.send_instcreation(
                        self._new_instance if self._new_instance is not None
                                           else self._old_instance,
                        fltr_id)
            else:
                LOG().debug("sending instance modification indication for job"
                        " %s with ID: %s", self._job, fltr_id)
                indication_manager.send_instmodification(
                        self._old_instance, self._new_instance,
                        fltr_id)

class JobManager(threading.Thread):
    """
    Separate thread for managing queue of jobs requested by client.
    There are three kinds of jobs, that are handled differently:

      * asynchronous - kept in ``_async_jobs`` dictionary until job is
            deleted by request or it expires;
            no reply is sent to client upon job's completion
      * synchronous  - reply is sent to client after job's completion;
            no reference to the job is kept afterwards
      * job control  - they are not enqueued in ``_job_queue`` for
            :py:class:`YumWorker` to process, but are handled directly and in
            the *FIFO* order

    Both asynchronous and synchronous jobs are enqueued in ``_job_queue`` and
    will be sent to :py:class:`~lmi.software.yumdb.process.YumWorker` for
    processing. It's a priority queue sorting jobs by their priority.
    :py:class:`~lmi.software.yumdb.sessionmanager.SessionManager` is the actual
    object managing ``YumWorker`` process. All jobs are passed to it.
    """
    #: enumeration of actions, that may be enqueued in calendar
    ACTION_REMOVE = 0

    #: names for actions defined above, their values are indexes to this list
    ACTION_NAMES = ['remove']

    def __init__(self):
        threading.Thread.__init__(self, name="JobManager")

        #: Session manager shall be accessed via property ``session_manager``.
        #: It's instantiated when first needed.
        self._sessionmgr = None

        #: lock for critical access to _sessionmgr, _calendar,
        #: _async_jobs and _job_queue
        self._job_lock = threading.RLock()
        #: priority queue of jobs that are processed by YumWorker
        self._job_queue = []
        #: (time, jobid, action)
        self._calendar = []
        #: {jobid : job}
        self._async_jobs = {}
        #: Condition object being wait upon in the main loop of this thread.
        #: It's notified either when the new job request comes or when the
        #: YumWorker replies.
        self._job_pending = threading.Condition(self._job_lock)

        #: used to guard access to _finished set
        self._reply_lock = threading.Lock()
        #: used to wait for job to be processed and received
        self._reply_cond = threading.Condition(self._reply_lock)
        #: job ids are kept here for each finished job
        self._finished = set()

    # *************************************************************************
    # Private methods
    # *************************************************************************
    @cmpi_logging.trace_method
    def _control_job(self, job):
        """
        Function dispatching job to handler for particular
        :py:class:`~lmi.software.yumdb.jobs.YumJob` subclass.

        .. note::
            This can be called only from client thread.
        """
        try:
            handler = {
                # these are from YumDB client
                jobs.YumJobGetList     : self._handle_get_list,
                jobs.YumJobGet         : self._handle_get,
                jobs.YumJobGetByName   : self._handle_get_by_name,
                jobs.YumJobSetPriority : self._handle_set_priority,
                jobs.YumJobReschedule  : self._handle_reschedule,
                jobs.YumJobUpdate      : self._handle_update,
                jobs.YumJobDelete      : self._handle_delete,
                jobs.YumJobTerminate   : self._handle_terminate,
            }[job.__class__]
            LOG().info("processing control job %s", str(job))
        except KeyError:
            raise errors.UnknownJob("No handler for job \"%s\"." %
                    job.__class__.__name__)
        return handler(**job.job_kwargs)

    @cmpi_logging.trace_method
    def _enqueue_job(self, job):
        """
        Insert incoming job into :py:attr:`_job_queue` or handle it directly if
        it's control job.

        .. note::
            This can be called only from client thread.
        """
        if isinstance(job, jobs.YumJobControl):
            result = job.RESULT_SUCCESS
            job.start()
            try:
                data = self._control_job(job)
            except Exception:   #pylint: disable=W0703
                result = job.RESULT_ERROR
                data = sys.exc_info()
                data = (data[0], data[1], traceback.format_tb(data[2]))
                LOG().exception("control job %s failed", job)
            self._finish_job(job, result, data)
        else:
            LOG().debug('job %s enqued for YumWorker to handle', job)
            heapq.heappush(self._job_queue, job)
            if getattr(job, 'async', False) is True:
                ind = self._prepare_indication_for(job, JM.IND_JOB_CREATED)
                self._async_jobs[job.jobid] = job
                ind.send()
            self._job_pending.notify()

    @cmpi_logging.trace_method
    def _schedule_event(self, after, jobid, action):
        """
        Enqueue event into calendar. Event consists of *time*, *jobid* and
        *action*.

        :param float after: Number of seconds from now until the action is
            shall be triggered.
        :param int jobid: Id of job which is a subject of action execution.
        :param int action: Value of action to do.

        .. note::
            This can be called only from client thread.
        """
        schedule_at = time.time() + after
        for (sched, jid, act) in self._calendar:
            if jid == jobid and act == action:
                if sched <= schedule_at:    # same event already scheduled
                    return
                # schedule it for early time
                LOG().debug('rescheduling action %s on job %d to take place'
                        ' after %d seconds (instead of %d)',
                        self.ACTION_NAMES[action], jid, after,
                        sched - schedule_at + after)
                self._calendar.remove((sched, jid, act))
                self._calendar.append((schedule_at, jid, act))
                heapq.heapify(self._calendar)
                return
        LOG().debug('scheduling action %s on job %d to take place after '
                ' %d seconds', self.ACTION_NAMES[action], jobid, after)
        heapq.heappush(self._calendar, (schedule_at, jobid, action))

    @cmpi_logging.trace_method
    def _run_event(self, jobid, action):
        """
        Process event from calendar.

        :param int jobid: Id of job that is a subject of action.
        :param int action: One of predefined actions to be executed on
            particular job.
        """
        LOG().info('running action %s on job(id=%d)',
                self.ACTION_NAMES[action], jobid)
        if action == self.ACTION_REMOVE:
            with self._job_lock:
                del self._async_jobs[jobid]
        else:
            msg = "unsupported action: %s" % action
            raise ValueError(msg)

    @cmpi_logging.trace_method
    def _prepare_indication_for(self, job, *args, **kwargs):
        """
        Convenience method making indication sender context manager.

        :param list args: Positional arguments passed after *job* to
            the constructor of :py:class:`JobIndicationSender`.
        :param dictionary kwargs: Its keyword arguments.
        :returns: Context manager for snapshoting and sending job replated
            indications.
        :rtype: :py:class:`JobIndicationSender`
        """
        return JobIndicationSender(IndicationManager.get_instance(), job,
            *args, **kwargs)

    @property
    def session_manager(self):
        """
        Accessor of session manager. Object is instantiated upon first access.

        :rtype: :py:class:`lmi.software.yumdb.sessionmanager.SessionManager`
        """
        with self._job_lock:
            if self._sessionmgr is None:
                self._sessionmgr = SessionManager(self._job_pending)
                self._sessionmgr.start()
            return self._sessionmgr

    # *************************************************************************
    # Job handlers
    # *************************************************************************
    @job_handler()
    def _handle_get(self, job):
        """ Control job handler returning job object requested. """
        return job

    @job_handler(False)
    def _handle_get_list(self):
        """ Control job handler returning list of all asynchronous jobs. """
        with self._job_lock:
            return sorted(self._async_jobs.values())

    @job_handler(False)
    def _handle_get_by_name(self, target):
        """
        Control job handler returning job object requested by its name.

        :param string target: Name of job to find. ``Name`` property of
            available asynchronous jobs is queried.
        """
        for job in self._async_jobs.values():
            if 'name' in job.metadata and target == job.metadata['name']:
                return job
        raise errors.JobNotFound(target)

    @job_handler()
    def _handle_set_priority(self, job, new_priority):
        """
        Control job handler modifying job's priority and updating its position
        in queue.

        :returns: Modified job object.
        :rtype: :py:class:`~lmi.software.yumdb.jobs.YumJob`
        """
        if not isinstance(new_priority, (int, long)):
            raise TypeError('priority must be an integer')
        if job.priority != new_priority:
            ind = self._prepare_indication_for(job)
            job.update(priority=new_priority)
            if job in self._job_queue:
                heapq.heapify(self._job_queue)
            ind.send(True)
        return job

    @job_handler()
    def _handle_reschedule(self, job,
            delete_on_completion,
            time_before_removal):
        """
        Control job handler rescheduling job's deletion.

        :param job: Affected job object.
        :type job: :py:class:`~lmi.software.yumdb.jobs.YumJob`
        :param boolean delete_on_completion: Whether the job shall be
            automatically removed after its completion. The
            *time_before_removel* then comes into play.
        :param float time_before_removal: Number of seconds to wait before
            job's automatic removal.
        """
        if (   job.delete_on_completion == delete_on_completion
           and job.time_before_removal == time_before_removal):
            return
        if job.finished and job.delete_on_completion:
            for i, event in enumerate(self._calendar):
                if event[1] == job.jobid and event[2] == self.ACTION_REMOVE:
                    del self._calendar[i]
                    heapq.heapify(self._calendar)
                    break
        ind = self._prepare_indication_for(job)
        if delete_on_completion:
            schedule_at = time_before_removal
            if job.finished:
                schedule_at = job.finished + schedule_at - time.time()
            self._schedule_event(schedule_at, job.jobid, self.ACTION_REMOVE)
        job.delete_on_completion = delete_on_completion
        job.time_before_removal = time_before_removal
        ind.send(True)
        return job

    @job_handler()
    def _handle_update(self, job, data):
        """
        Control job handler updating any job metadata.

        :param dictionary data: Job attributes with associated values that
            shall be assigned to particular *job*.
        """
        ind = self._prepare_indication_for(job)
        job.update(**data)
        ind.send(True)
        return job

    @job_handler()
    def _handle_delete(self, job):
        """
        Control job handler deleting finished asynchronous job. If the *job* is
        not yet finished, the
        :py:exc:`lmi.software.yumdb.errors.InvalidJobState` is raised.
        """
        if not job.finished:
            raise errors.InvalidJobState(
                    'can not delete unfinished job "%s"' % job)
        try:
            self._job_queue.remove(job)
            heapq.heapify(self._job_queue)
            LOG().debug('job "%s" removed from queue', job)
        except ValueError:
            LOG().debug('job "%s" not started and not enqueued', job)
        del self._async_jobs[job.jobid]
        return job

    @job_handler()
    def _handle_terminate(self, job):
        """
        Control job handler terminating not started job.
        """
        if job.started and not job.finished:
            raise errors.InvalidJobState('can not kill running job "%s"' % job)
        if job.finished:
            raise errors.InvalidJobState('job "%s" already finished' % job)
        self._job_queue.remove(job)
        heapq.heapify(self._job_queue)
        ind = self._prepare_indication_for(job)
        job.finish(result=job.RESULT_TERMINATED)
        ind.send(True)
        LOG().info('terminated not started job "%s"', job)
        return job

    # *************************************************************************
    # Public methods
    # *************************************************************************
    @cmpi_logging.trace_method
    def _finish_job(self, job, result, result_data):
        """
        This should be called on any processed job.

        If the *job* is synchronous, reply is send at once. Otherwise the
        result is stored for later client's query in the job itself, and
        indications are sent.

        .. note::
            This shall be called only from :py:meth:`run` method.

        :param int result: Result code that will be stored in a *job*.
        :param result_data: Resulting data object that is to be stored in a
            *job*.
        :returns: Job object.
        """
        if job.state != job.RUNNING:
            raise errors.InvalidJobState(
                    'can not finish not started job "%s"' % job)
        if getattr(job, 'async', False):
            ind = self._prepare_indication_for(job,
                    (JM.IND_JOB_CHANGED, JM.IND_JOB_PERCENT_UPDATED))
        job.finish(result, result_data)
        if getattr(job, 'async', False):
            if job.delete_on_completion:
                default = Configuration.get_instance().get_safe(
                            'Log', 'MinimumTimeBeforeRemoval', float)
                schedule_at = max(job.time_before_removal, default)
                self._schedule_event(schedule_at, job.jobid,
                        self.ACTION_REMOVE)
            if result == job.RESULT_SUCCESS:
                ind.add_indication_ids(JM.IND_JOB_SUCCEEDED)
            elif result == job.RESULT_ERROR:
                ind.add_indication_ids(JM.IND_JOB_FAILED)
            ind.send(True)
        with self._reply_lock:
            self._finished.add(job.jobid)
            self._reply_cond.notifyAll()
        return job

    @cmpi_logging.trace_method
    def _get_job(self):
        """
        Pop the first job enqueued by client out of :py:attr:`_job_queue`.
        Job is started and returned.

        :returns: Job object. If the queue is empty, ``None`` is returned.
        :rtype: :py:class:`~lmi.software.yumdb.jobs.YumJob`
        """
        if self._job_queue:
            job = heapq.heappop(self._job_queue)
            if getattr(job, "async", False):
                ind = self._prepare_indication_for(job,
                        (JM.IND_JOB_CHANGED, JM.IND_JOB_PERCENT_UPDATED))
                job.start()
                ind.send(True)
            else:
                job.start()
            return job

    @cmpi_logging.trace_method
    def process(self, job):
        """
        Enqueue given job and block until it's processed.

        :returns: Modified job object with assigned result.
        :rtype: :py:class:`~lmi.software.yumdb.jobs.YumJob`
        """
        with self._job_lock:
            self._enqueue_job(job)
        if getattr(job, 'async', False):
            return job.jobid
        with self._reply_lock:
            while job.jobid not in self._finished:
                self._reply_cond.wait()
            return job

    @cmpi_logging.trace_method
    def _clean_up(self):
        """
        Release the session manager with its separated process.
        This shall be called from the main loop of this thread.
        """
        with self._job_lock:
            if self._sessionmgr is not None:
                self._sessionmgr.join()
                self._sessionmgr = False

    @cmpi_logging.trace_method
    def begin_session(self):
        """
        Nest into a session. This means: lock the yum database for exclusive
        use until the top-level session ends. The opposite method,
        :py:meth:`end_session` needs to be called same number of times in order
        for yum database to be released.
        """
        return self.session_manager.begin_session()

    @cmpi_logging.trace_method
    def end_session(self):
        """
        Emerge from session. When the last session ends, yum database is
        unlocked.

        .. see::
            :py:meth:`begin_session`
        """
        return self.session_manager.end_session()

    def run(self):
        """
        Main lool of *JobManager*'s thread. It works with just one job object
        at once until it's processed. It consumes enqueued jobs in
        :py:attr:`_job_queue` and passes them to session manager object. It
        also triggers actions scheduled in calendar.
        """
        LOG().info("%s thread started", self.name)
        # This points to the currect job being processed by session manager.
        # If no job is being processed, it's ``None``.
        job = None
        with self._job_lock:
            while True:
                timeout = None

                # check calendar for scheduled events
                if self._calendar:
                    timeout = self._calendar[0][0] - time.time()
                while timeout is not None and timeout <= 0:
                    _, jobid, action = heapq.heappop(self._calendar)
                    self._run_event(jobid, action)
                    if self._calendar:
                        timeout = self._calendar[0][0] - time.time()
                    else:
                        timeout = None

                # handle processed job if any
                if job:
                    if self.session_manager.got_reply:
                        _, result, data = self.session_manager.pop_reply()
                        self._finish_job(job, result, data)
                        if isinstance(job, jobs.YumShutDown):
                            break
                        job = None

                # check newly enqueued jobs
                if job is None:
                    job = self._get_job()
                    if job:
                        self.session_manager.process(job, sync=False)

                # wait for any event
                LOG().debug('waiting on input queue for job%s',
                    (' with timeout %s' % timeout) if timeout else '')
                self._job_pending.wait(timeout=timeout)

        self._clean_up()
        LOG().info('%s thread terminating', self.name)




#!/usr/bin/env python
#
# Copyright (C) 2012-2013 Red Hat, Inc.  All rights reserved.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#
# Authors: Michal Minar <miminar@redhat.com>
#
"""
Unit tests for ``LMI_SoftwareInstallationServiceAffectsElement`` provider.
"""

import unittest

import swbase

class TestSoftwareInstallationServiceAffectsElement(swbase.SwTestCase):
    """
    Basic cim operations test for
    ``LMI_SoftwareInstallationServiceAffectsElement``.
    """

    CLASS_NAME = "LMI_SoftwareInstallationServiceAffectsElement"
    KEYS = ("AffectedElement", "AffectingElement")

    def make_op(self, pkg=None):
        """
        :returns Object path of
            ``LMI_SoftwareInstallationServiceAffectsElement``.
        :rtype: :py:class:`lmi.shell.LMIInstanceName`
        """
        op = self.cim_class.new_instance_name({
            "AffectingElement" : self.ns.LMI_SoftwareInstallationService. \
                new_instance_name({
                    "Name" : 'LMI:LMI_SoftwareInstallationService',
                    "SystemName" : self.SYSTEM_NAME,
                    "SystemCreationClassName" : self.system_cs_name,
                    "CreationClassName" : "LMI_SoftwareInstallationService"
                }),
            })
        if pkg is not None:
            op.wrapped_object["AffectedElement"] = \
                    self.ns.LMI_SoftwareIdentity.new_instance_name({
                        "InstanceID" : 'LMI:LMI_SoftwareIdentity:' + pkg.nevra
                    }).wrapped_object
        else:
            op.wrapped_object["AffectedElement"] = \
                    self.system_iname.wrapped_object
        return op

    @swbase.test_with_repos(**{
        'updates' : True,
        'misc' : True,
        'stable' : False,
        'updates-testing': False
    })
    @swbase.test_with_packages(**{
        'updates#pkg1' : True,
        'pkg2' : False,
        'pkg3' : False,
        'pkg4' : False
    })
    def test_get_instance_on_available_packages(self):
        """
        Test ``GetInstance()`` call on
        ``LMI_SoftwareInstallationServiceAffectsElement``.
        """
        for repoid in ('updates', 'misc'):
            repo = self.get_repo(repoid)
            for pkg in repo.packages:
                objpath = self.make_op(pkg)
                inst = objpath.to_instance()
                self.assertNotEqual(inst, None,
                        'Package "%s" is affected by installation service'
                        % pkg)
                self.assertCIMNameEqual(inst.path, objpath,
                        "Object paths should match for package %s" % pkg)
                self.assertEqual(set(inst.path.key_properties()),
                        set(self.KEYS))

    @swbase.test_with_repos(misc=False)
    @swbase.test_with_packages(**{
        'misc#pkg2' : True,
    })
    def test_get_instance_with_not_available_package(self):
        """
        Test ``GetInstance()`` call on
        ``LMI_SoftwareInstallationServiceAffectsElement`` with not available
        but installed package.
        """
        repo = self.get_repo('misc')
        for pkg in repo.packages:
            objpath = self.make_op(pkg)
            inst = objpath.to_instance()
            self.assertEqual(inst, None, 'package should not be affected')

    def test_get_instance_on_computer_system(self):
        """
        Test ``GetInstance()`` call on
        ``LMI_SoftwareInstallationServiceAffectsElement`` on computer system.
        """
        objpath = self.make_op()
        inst = objpath.to_instance()
        self.assertNotEqual(inst, None, "computer system shall be affected")
        self.assertCIMNameEqual(inst.path, objpath)

    @swbase.test_with_repos(**{
        'stable' : True,
        'updates': False,
        'updates-testing': True,
        'misc' : False
    })
    @swbase.test_with_packages(**{
        'stable#pkg1' : True,
        'updates#pkg2' : True,
        'pkg3' : False,
        'pkg4' : False
    })
    def test_enum_instances(self):
        """
        Test ``EnumInstances()`` call on
        ``LMI_SoftwareInstallationServiceAffectsElement``.
        """
        insts = self.cim_class.instances()
        self.assertGreater(len(insts), 0)
        for inst in insts:
            self.assertEqual(set(inst.path.key_properties()), set(self.KEYS))
            for key in self.KEYS:
                self.assertCIMNameEqual(
                        getattr(inst, key), getattr(inst.path, key))

        nevra_set = set(i.AffectedElement.InstanceID[
                len('LMI:LMI_SoftwareIdentity:'):] for i in insts
                if i.AffectedElement.classname == 'LMI_SoftwareIdentity')
        # There is also computer system associated
        self.assertEqual(len(nevra_set) + 1, len(insts))
        for repo in self.repodb.values():
            if not repo.repoid in (
                    'openlmi-sw-test-repo-stable',
                    'openlmi-sw-test-repo-updates-testing'):
                continue
            for pkg in repo.packages:
                self.assertIn(pkg.nevra, nevra_set)
                nevra_set.remove(pkg.nevra)
        self.assertEqual(len(nevra_set), 0)

        objpath = self.make_op()
        cs = None # computer system instance name
        for i in insts:
            if i.AffectedElement.classname == objpath.AffectedElement.classname:
                cs = i.AffectedElement
                break
        self.assertNotEqual(cs, None,
                "computer system is one of affected elements")
        self.assertCIMNameEqual(cs, objpath.AffectedElement)

    @swbase.test_with_repos(**{
        'stable' : True,
        'updates': False,
        'updates-testing': True,
        'misc' : False
    })
    def test_enum_instance_names(self):
        """
        Test ``EnumInstanceNames()`` call on
        ``LMI_SoftwareInstallationServiceAffectsElement``.
        """
        inames = self.cim_class.instance_names()
        self.assertGreater(len(inames), 1)
        objpath = self.make_op(self.get_repo('stable')['pkg1'])
        for iname in inames:
            self.assertEqual(iname.namespace, 'root/cimv2')
            self.assertEqual(iname.classname, self.CLASS_NAME)
            self.assertEqual(sorted(iname.key_properties()), sorted(self.KEYS))
            self.assertCIMNameEqual(objpath.AffectingElement,
                    iname.AffectingElement)
        nevra_set = set(i.AffectedElement.InstanceID[
                len('LMI:LMI_SoftwareIdentity:'):] for i in inames
                if i.AffectedElement.classname == 'LMI_SoftwareIdentity')
        # There is also computer system associated
        self.assertEqual(len(nevra_set) + 1, len(inames))
        for repo in self.repodb.values():
            if not repo.repoid in (
                    'openlmi-sw-test-repo-stable',
                    'openlmi-sw-test-repo-updates-testing'):
                continue
            for pkg in repo.packages:
                self.assertIn(pkg.nevra, nevra_set)
                nevra_set.remove(pkg.nevra)
        self.assertEqual(len(nevra_set), 0)

        objpath = self.make_op()
        cs = None # computer system instance name
        for i in inames:
            if i.AffectedElement.classname == objpath.AffectedElement.classname:
                cs = i.AffectedElement
                break
        self.assertNotEqual(cs, None,
                "computer system is one of affected elements")
        self.assertCIMNameEqual(cs, objpath.AffectedElement)

    def test_get_service_computer_system(self):
        """
        Try to get computer system associated with service.
        """
        objpath = self.make_op()
        service = objpath.AffectingElement.to_instance()
        self.assertNotEqual(service, None,
                "GetInstance call succeeds on service")
        refs = service.associator_names(
                AssocClass='LMI_SoftwareInstallationServiceAffectsElement',
                ResultClass=self.system_cs_name,
                Role="AffectingElement",
                ResultRole="AffectedElement")
        self.assertEqual(len(refs), 1)
        cs = refs[0]
        self.assertCIMNameEqual(cs, objpath.AffectedElement)

    @swbase.test_with_repos(**{
        'updates' : True,
        'misc' : True,
        'stable' : False,
        'updates-testing' : False
    })
    def test_get_service_package_names(self):
        """
        Try to get all package names associated with installation service.
        """
        objpath = self.make_op()
        service = objpath.AffectingElement.to_instance()
        self.assertNotEqual(service, None,
                "GetInstance call succeeds on service")
        refs = service.associator_names(
                AssocClass='LMI_SoftwareInstallationServiceAffectsElement',
                ResultClass="LMI_SoftwareIdentity",
                Role="AffectingElement",
                ResultRole="AffectedElement")
        self.assertGreater(len(refs), 0)

        for ref in refs:
            self.assertEqual(ref.namespace, 'root/cimv2')
            self.assertEqual(ref.classname, "LMI_SoftwareIdentity")
            self.assertEqual(ref.key_properties(), ["InstanceID"])
            self.assertTrue(ref.InstanceID.startswith(
                "LMI:LMI_SoftwareIdentity:"))

        nevra_set = set(   i.InstanceID[len('LMI:LMI_SoftwareIdentity:'):]
                       for i in refs)
        for repo in self.repodb.values():
            if not repo.repoid in (
                    'openlmi-sw-test-repo-updates',
                    'openlmi-sw-test-repo-misc'):
                continue
            for pkg in repo.packages:
                self.assertIn(pkg.nevra, nevra_set)
                nevra_set.remove(pkg.nevra)
        self.assertEqual(len(nevra_set), 0)

    def test_get_computer_system_service(self):
        """
        Try to get software installation service associated with computer
        system.
        """
        objpath = self.make_op()
        refs = objpath.AffectedElement.to_instance().associator_names(
                AssocClass=self.CLASS_NAME,
                Role="AffectedElement",
                ResultRole="AffectingElement",
                ResultClass='LMI_SoftwareInstallationService')
        self.assertEqual(len(refs), 1)
        ref = refs[0]
        self.assertCIMNameEqual(objpath.AffectingElement, ref)

    @swbase.test_with_repos('stable')
    @swbase.test_with_packages(**{
        'stable#pkg1' : False
    })
    def test_get_package_installation_service(self):
        """
        Try to get installation service associated to package.
        """
        pkg = self.get_repo('stable')['pkg1']
        objpath = self.make_op(pkg)
        refs = objpath.AffectedElement.to_instance().associator_names(
                AssocClass=self.CLASS_NAME,
                Role="AffectedElement",
                ResultRole="AffectingElement",
                ResultClass='LMI_SoftwareInstallationService')
        self.assertEqual(len(refs), 1)
        ref = refs[0]
        self.assertCIMNameEqual(objpath.AffectingElement, ref)

    @swbase.test_with_repos(**{
        'stable' : True,
        'updates' : False,
        'updates-testing' : False,
        'misc' : False
    })
    @swbase.test_with_packages(**{
        'stable#pkg1' : True,
        'pkg2' : False,
        'stable#pkg3' : True,
        'pkg3' : False
    })
    def test_service_affected_elements(self):
        """
        Try to get all affected elements associated with installation service.
        """
        pkg = self.get_repo('stable')['pkg1']
        objpath = self.make_op(pkg)
        service = objpath.AffectingElement.to_instance()
        self.assertNotEqual(service, None)
        refs = service.associators(
                AssocClass=self.CLASS_NAME,
                Role="AffectingElement",
                ResultRole="AffectedElement")
        # number of packages in stable (4) + computer system (1)
        self.assertEqual(len(refs), 5)
        nevra_set = set()
        for ref in refs:
            if ref.classname == self.system_cs_name:
                self.assertCIMNameEqual(ref.path, self.system_iname)
            else:
                self.assertEqual(ref.classname, 'LMI_SoftwareIdentity')
                self.assertEqual(ref.namespace, 'root/cimv2')
                nevra_set.add(ref.ElementName)
        for pkg in self.get_repo('stable').packages:
            self.assertIn(pkg.nevra, nevra_set)
            nevra_set.remove(pkg.nevra)
        self.assertEqual(len(nevra_set), 0)

def suite():
    """For unittest loaders."""
    return unittest.TestLoader().loadTestsFromTestCase(
            TestSoftwareInstallationServiceAffectsElement)

if __name__ == '__main__':
    unittest.main()



# -*- encoding: utf-8 -*-
# Software Management Providers
#
# Copyright (C) 2012-2013 Red Hat, Inc.  All rights reserved.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#
# Authors: Michal Minar <miminar@redhat.com>
#
"""
Module holding the code of separate process accessing the YUM API.
"""

import errno
from functools import wraps
import logging
from multiprocessing import Process
import os
import Queue as TQueue   # T as threaded
import re
import sys
import time
import traceback
from urlgrabber.grabber import default_grabber
import weakref
import yum

from lmi.providers import cmpi_logging
from lmi.software import util
from lmi.software.yumdb import errors
from lmi.software.yumdb import jobs
from lmi.software.yumdb import packageinfo
from lmi.software.yumdb import packagecheck
from lmi.software.yumdb import repository
from lmi.software.yumdb.util import is_pkg_installed
from lmi.software.yumdb.util import setup_logging

# Global variable which gets its value after the start of YumWorker process.
LOG = None

class ShutDownWorker(Exception):
    """ Raised when shut down job is received. """
    pass

# *****************************************************************************
# Utilities
#  ****************************************************************************
def _get_package_filter_function(filters, exact_match=True):
    """
    :param filters: (``dict``) Dictionary with keys of package property
        names and values of their desired values.
    :param exact_match: (``bool``) Whether the ``name`` should be checked
        for exact match or for presence in package's name or summary
        strings.
    :rtype: (``function``) A function used to filter list of packages.
    """
    if not isinstance(filters, dict):
        raise TypeError("filters must be a dictionary")

    filters = dict((k, value) for k, value in filters.items()
                              if value is not None)

    match = None
    if "nevra" in filters:
        match = util.RE_NEVRA.match(filters["nevra"])
    elif "envra" in filters:
        match = util.RE_ENVRA.match(filters["envra"])
    if match is not None:
        for attr in ("name", "epoch", "version", "release", "arch"):
            match_attr = attr
            filters[attr] = match.group(match_attr)
        filters.pop('nevra', None)
        filters.pop('envra', None)
    elif "evra" in filters:
        for prop_name in ("epoch", "version", "release", "epoch"):
            filters.pop(prop_name, None)

    filter_list = []
    # properties are sorted by their filtering ability
    # (the most unprobable property, that can match, comes first)
    for prop_name in ("evra", "name", "version", "epoch",
            "release", "repoid", "arch"):
        if not prop_name in filters:
            continue
        if prop_name == 'name' and not exact_match and 'name' in filters:
            continue
        filter_list.append((prop_name, filters.pop(prop_name)))
    if not exact_match and 'name' in filters:
        re_name = re.compile(re.escape(filters['name']))
        def _cmp_props(pkg):
            """ :rtype: (``bool``) Does pkg matche properies filter? """
            if re_name.search(pkg.name) or re_name.search(pkg.summary):
                return all(getattr(pkg, p) == v for p, v in filter_list)
            return False
    else:
        def _cmp_props(pkg):
            """ :rtype: (``bool``) Does pkg matche properies filter? """
            return all(getattr(pkg, p) == v for p, v in filter_list)
    return _cmp_props

class RepoFilterSetter(object):
    """
    A context manager, that will set a repository filter lasting
    as long as the object itself.
    """
    def  __init__(self, yum_base, include_repos=None, exclude_repos=None):
        if not isinstance(yum_base, yum.YumBase):
            raise TypeError("yum_base must be a YumBase instance")
        self._yum_base = yum_base
        self._include = include_repos
        self._exclude = exclude_repos
        # after __enter__ this will be dictionary containing (
        # repoid, enabled) pairs
        self._prev_states = None

    def __enter__(self):
        self._prev_states = {   r.id: r.enabled
                            for r in self._yum_base.repos.repos.values()}
        if isinstance(self._exclude, (list, tuple, set)):
            exclude = ",".join(self._exclude)
        else:
            exclude = self._exclude
        # set of repositories, that were affected
        repos = set()
        if exclude:
            repos.update(self._yum_base.repos.disableRepo(exclude))
            LOG.info('disabling repositories: [%s]', ", ".join(repos))
        if isinstance(self._include, (list, tuple, set)):
            include = ",".join(self._include)
        else:
            include = self._include
        if include:
            affected = self._yum_base.repos.enableRepo(include)
            LOG.info('enabling repositories: [%s]', ", ".join(affected))
            repos.update(affected)
        for repoid, prev_enabled in self._prev_states.items():
            if (  repoid not in repos
               or (  bool(prev_enabled)
                  is bool(self._yum_base.repos.getRepo(repoid).enabled))):
                # keep only manipulated repositories
                del self._prev_states[repoid]
        if len(self._prev_states):
            for repoid in (r for r, v in self._prev_states.items() if v):
                self._yum_base.pkgSack.sacks.pop(repoid, None)
            self._yum_base.repos.populateSack()
        return self

    def __exit__(self, exc_type, exc_value, exc_tb):
        # restore previous repository states
        if len(self._prev_states):
            LOG.info('restoring repositories: [%s]',
                    ", ".join(self._prev_states.keys()))
            for repoid, enabled in self._prev_states.items():
                repo = self._yum_base.repos.getRepo(repoid)
                if enabled:
                    repo.enable()
                else:
                    repo.disable()
            for repoid in (r for r, v in self._prev_states.items() if not v):
                self._yum_base.pkgSack.sacks.pop(repoid, None)
            self._yum_base.repos.populateSack()

# *****************************************************************************
# Decorators
# *****************************************************************************
def _needs_database(method):
    """
    Decorator for YumWorker job handlers, that need to access the yum database.
    It ensures, that database is initialized and locks it in case, that
    no session is active.
    """
    logged = cmpi_logging.trace_method(method, frame_level=2)
    @wraps(method)
    def _wrapper(self, *args, **kwargs):
        """
        Wrapper for the job handler method.
        """
        created_session = False
        self._init_database()
        if self._state != YumWorker.STATE_LOCKED:
            created_session = True
            self._lock_database()
            self._check_repository_configs()
        try:
            result = logged(self, *args, **kwargs)
            return result
        finally:
            if created_session is True:
                self._unlock_database()
    return _wrapper

# *****************************************************************************
# Classes
# *****************************************************************************
class YumWorker(Process):
    """
    The main process, that works with YUM API. It has two queues, one
    for input jobs and second for results.

    Jobs are dispatched by their class names to particular handler method.

    It can process just one job at once. It has following states:

        ``STATE_IDLE``
            Initial state where no information is cached. Next state
            is ``STATE_LOCKED``.
        ``STATE_LOCKED``
            Database is locked, and :py:class:`yum.YumBase` is initialied.
            Next state is ``STATE_UNLOCKED``.
        ``STATE_UNLOCKED``
            Database is unlocked. Yum objects are cached. Next state is
            any of above ones.
    """

    STATE_IDLE, STATE_LOCKED, STATE_UNLOCKED = range(3)

    def __init__(self,
            queue_in,
            queue_out,
            yum_kwargs=None):
        Process.__init__(self, name="YumWorker")
        self._queue_in = queue_in
        self._queue_out = queue_out
        self._state = YumWorker.STATE_IDLE

        if yum_kwargs is None:
            yum_kwargs = {}

        self._yum_kwargs = yum_kwargs
        self._yum_base = None

        self._pkg_cache = None
        # contains (repoid, time_stamp_of_config_file)
        # plus (/repos/dir, ...) for each repo config directory
        self._repodir_mtimes = {}

    # *************************************************************************
    # Private methods
    # *************************************************************************
    @cmpi_logging.trace_method
    def _init_database(self):
        """
        Initializes yum base object, when it does no exists.
        And updates the cache (when out of date).
        """
        if self._yum_base is None:
            LOG.info("creating YumBase with kwargs=(%s)",
                ", ".join((   "%s=%s"%(k, str(v))
                          for k, v in self._yum_kwargs.items())))
            self._yum_base = yum.YumBase(**self._yum_kwargs)

    @cmpi_logging.trace_method
    def _free_database(self):
        """
        Release the yum base object to safe memory.
        """
        LOG.info("freing database")
        self._pkg_cache.clear()
        self._yum_base = None
        self._state = YumWorker.STATE_IDLE

    @cmpi_logging.trace_method
    def _lock_database(self):
        """
        Only one process is allowed to work with package database at given time.
        That's why we lock it.

        Try to lock it in loop, until success.
        """
        assert self._state != YumWorker.STATE_LOCKED
        while True:
            try:
                LOG.info("trying to lock database")
                self._yum_base.doLock()
                self._state = YumWorker.STATE_LOCKED
                LOG.info("successfully locked up")
                break
            except yum.Errors.LockError as exc:
                LOG.warn("failed to lock")
                if exc.errno in (errno.EPERM, errno.EACCES, errno.ENOSPC):
                    LOG.error("can't create lock file")
                    raise errors.DatabaseLockError("Can't create lock file.")
            timeout = util.Configuration.get_instance().get_safe(
                    'Yum', 'LockWaitInterval', float)
            LOG.info("trying to lock again after %.1f seconds", timeout)
            time.sleep(timeout)

    @cmpi_logging.trace_method
    def _unlock_database(self):
        """
        The opposite to _lock_database() method.
        """
        if self._yum_base is not None:
            LOG.info("unlocking database")
            self._yum_base.closeRpmDB()
            del self._yum_base.history
            self._yum_base.doUnlock()
            self._state = YumWorker.STATE_UNLOCKED

    @cmpi_logging.trace_method
    def _get_job(self):
        """
        Wait for new job to process. If no job comes for a long time, free
        database to save memory.
        """
        while True:
            if self._state == YumWorker.STATE_UNLOCKED:
                try:
                    timeout = util.Configuration.get_instance().get_safe(
                            'Yum', 'FreeDatabaseTimeout', float)
                    return self._queue_in.get(timeout=timeout)
                except TQueue.Empty:
                    self._free_database()
            else:
                return self._queue_in.get()

    @cmpi_logging.trace_method
    def _transform_packages(self, packages,
            cache_packages=True,
            flush_cache=True):
        """
        Return instances of PackageInfo for each package in packages.
        Cache all the packages.
        @param packages list of YumAvailablePackage instances
        @param cache_packages whether to update cache with packages
        @param flush_cache whether to clear the cache before adding input
        packages; makes sense only with cachee_packages=True
        """
        if cache_packages is True and flush_cache is True:
            LOG.debug("flushing package cache")
            self._pkg_cache.clear()
        res = []
        for orig in packages:
            pkg = packageinfo.make_package_from_db(orig, self._yum_base.rpmdb)
            if cache_packages is True:
                self._pkg_cache[pkg.objid] = orig
            res.append(pkg)
        return res

    @cmpi_logging.trace_method
    def _cache_packages(self, packages, flush_cache=True, transform=False):
        """
        Store packages in cache and return them.
        @param flush_cache whether to clear the cache before adding new
        packages
        @param transform whether to return packages as PackageInfos
        @return either list of original packages or PackageInfo instances
        """
        if transform is True:
            return self._transform_packages(packages, flush_cache=flush_cache)
        if flush_cache is True:
            LOG.debug("flushing package cache")
            self._pkg_cache.clear()
        for pkg in packages:
            self._pkg_cache[id(pkg)] = pkg
        return packages

    @cmpi_logging.trace_method
    def _lookup_package(self, pkg):
        """
        Lookup the original package in cache.
        If it was garbage collected already, make new query to find it.
        @return instance of YumAvailablePackage
        """
        if not isinstance(pkg, packageinfo.PackageInfo):
            raise TypeError("pkg must be instance of PackageInfo")
        LOG.debug("looking up yum package %s with id=%d",
            pkg, pkg.objid)
        try:
            result = self._pkg_cache[pkg.objid]
            LOG.debug("lookup successful")
        except KeyError:
            LOG.warn("lookup of package %s with id=%d failed, trying"
                " to query database", pkg, pkg.objid)
            result = self._handle_filter_packages(
                    'installed' if pkg.installed else 'available',
                    allow_duplicates=False,
                    sort=False,
                    transform=False,
                    **pkg.key_props)
            if len(result) < 1:
                LOG.warn("package %s not found", pkg)
                raise errors.PackageNotFound(
                        "package %s could not be found" % pkg)
            result = result[0]
        return result

    @cmpi_logging.trace_method
    def _clear_repository_cache(self):
        """
        Clears the repository cache and their configuration directory
        last modification times.
        """
        if self._yum_base is not None:
            for repoid in self._yum_base.repos.repos.keys():
                self._yum_base.repos.delete(repoid)
            del self._yum_base.repos
            del self._yum_base.pkgSack
        self._repodir_mtimes.clear()

    @cmpi_logging.trace_method
    def _check_repository_configs(self):
        """
        Checks whether repository information is up to date with configuration
        files by comparing timestamps. If not, repository cache will be
        released.
        """
        dirty = False
        if self._repodir_mtimes:
            for repodir in self._yum_base.conf.reposdir:
                if (  os.path.exists(repodir)
                   and (   not repodir in self._repodir_mtimes
                       or  ( os.stat(repodir).st_mtime
                           > self._repodir_mtimes[repodir]))):
                    LOG.info("repository config dir %s changed", repodir)
                    dirty = True
                    break
            if not dirty:
                for repo in self._yum_base.repos.repos.values():
                    filename = repo.repofile
                    if (  not os.path.exists(filename)
                       or (  int(os.stat(filename).st_mtime)
                          > repo.repo_config_age)):
                        LOG.info('config file of repository "%s" changed',
                                repo.id)
                        dirty = True
                        break
        if dirty is True:
            LOG.info("repository cache is dirty, cleaning up ...")
            self._clear_repository_cache()
            self._yum_base.getReposFromConfig()
        if dirty is True or not self._repodir_mtimes:
            self._update_repodir_mtimes()

    @cmpi_logging.trace_method
    def _update_repodir_mtimes(self):
        """
        Updates the last modification times of repo configuration directories.
        """
        assert self._yum_base is not None
        for repodir in self._yum_base.conf.reposdir:
            if os.path.exists(repodir):
                self._repodir_mtimes[repodir] = os.stat(repodir).st_mtime

    @cmpi_logging.trace_method
    def _do_work(self, job):
        """
        Dispatcher of incoming jobs. Job is passed to the right handler
        depending on its class.
        """
        if not isinstance(job, jobs.YumJob):
            raise TypeError("job must be instance of YumJob")
        try:
            handler = {
                jobs.YumGetPackageList   : self._handle_get_package_list,
                jobs.YumFilterPackages   : self._handle_filter_packages,
                jobs.YumInstallPackage   : self._handle_install_package,
                jobs.YumRemovePackage    : self._handle_remove_package,
                jobs.YumUpdateToPackage  : self._handle_update_to_package,
                jobs.YumUpdatePackage    : self._handle_update_package,
                jobs.YumLock             : self._handle_lock,
                jobs.YumUnlock           : self._handle_unlock,
                jobs.YumCheckPackage     : self._handle_check_package,
                jobs.YumCheckPackageFile : self._handle_check_package_file,
                jobs.YumInstallPackageFromURI : \
                        self._handle_install_package_from_uri,
                jobs.YumGetRepositoryList : \
                        self._handle_get_repository_list,
                jobs.YumFilterRepositories : self._handle_filter_repositories,
                jobs.YumSetRepositoryEnabled : \
                        self._handle_set_repository_enabled
            }[job.__class__]
            LOG.info("processing job %s(id=%d)",
                job.__class__.__name__, job.jobid)
        except KeyError:
            if isinstance(job, jobs.YumShutDown):
                raise ShutDownWorker()
            LOG.error("No handler for job \"%s\"", job.__class__.__name__)
            raise errors.UnknownJob("No handler for job \"%s\"." %
                    job.__class__.__name__)
        return handler(**job.job_kwargs)

    @cmpi_logging.trace_method
    def _run_transaction(self, name):
        """
        Builds and runs the yum transaction and checks for errors.
        @param name of transaction used only in error description on failure
        """
        LOG.info("building transaction %s", name)
        (code, msgs) = self._yum_base.buildTransaction()
        if code == 1:
            LOG.error("building transaction %s failed: %s",
                name, "\n".join(msgs))
            raise errors.TransactionBuildFailed(
                "Failed to build \"%s\" transaction: %s" % (
                    name, "\n".join(msgs)))
        LOG.info("processing transaction %s", name)
        self._yum_base.processTransaction()
        self._yum_base.closeRpmDB()

    @cmpi_logging.trace_method
    def _main_loop(self):
        """
        This is a main loop called from run(). Jobs are handled here.
        It accepts a job from input queue, handles it and
        sends the result to output queue.

        It is terminated, when :py:class:`~lmi.software.yumdb.jobs.YumShutDown`
        object is received.
        """
        while True:
            job = self._get_job()
            result = jobs.YumJob.RESULT_SUCCESS
            try:
                data = self._do_work(job)
            except ShutDownWorker:
                break
            except Exception:    #pylint: disable=W0703
                result = jobs.YumJob.RESULT_ERROR
                # (type, value, traceback)
                data = sys.exc_info()
                # traceback is not pickable - replace it with formatted
                # text
                data = (data[0], data[1], traceback.format_tb(data[2]))
                LOG.exception("job %s failed", job)
            self._queue_out.put((job.jobid, result, data))
        self._unlock_database()
        self._free_database()

    @cmpi_logging.trace_method
    def _handle_lock(self):
        """ Handle lock request. """
        if self._state != YumWorker.STATE_LOCKED:
            self._init_database()
            self._lock_database()
            self._check_repository_configs()
        else:
            LOG.warn('requested lock while database is already locked')

    @cmpi_logging.trace_method
    def _handle_unlock(self):
        """
        Handler for unlock job.
        """
        if self._state == YumWorker.STATE_LOCKED:
            self._unlock_database()
        else:
            LOG.trace_warn('requested unlock on not locked database')

    @_needs_database
    def _handle_get_package_list(self, kind, allow_duplicates, sort,
            include_repos=None, exclude_repos=None, transform=True):
        """
        Handler for listing packages job.
        @param transform says, whether to return just a package abstractions
        or original ones
        @return [pkg1, pkg2, ...]
        """
        if kind == 'avail_notinst':
            what = 'available'
        elif kind == 'available':
            what = 'all'
        elif kind == 'avail_reinst':
            what = 'all'
        else:
            what = kind
        with RepoFilterSetter(self._yum_base, include_repos, exclude_repos):
            LOG.debug("calling YumBase.doPackageLists(%s, showdups=%s)",
                    what, allow_duplicates)
            pkglist = self._yum_base.doPackageLists(what,
                    showdups=allow_duplicates)
        if kind == 'all':
            result = pkglist.available + pkglist.installed
        elif kind == 'available':
            result = pkglist.available + pkglist.reinstall_available
        elif kind == 'avail_reinst':
            result = pkglist.reinstall_available
        else:   # get installed or available
            result = getattr(pkglist, what)
        if sort is True:
            result.sort()
        LOG.debug("returning %s packages", len(result))
        return self._cache_packages(result, transform=transform)

    @_needs_database
    def _handle_filter_packages(self, kind, allow_duplicates, sort,
            exact_match=True, include_repos=None, exclude_repos=None,
            transform=True, **filters):
        """
        Handler for filtering packages job.
        @return [pkg1, pkg2, ...]
        """
        if kind == 'all' \
               and filters.get('repoid', None) \
               and not include_repos and not exclude_repos:
            # If repoid is given, we are interested only in available
            # packages, not installed. With kind == 'all' we are not able to
            # check repoid of installed package (which may be available).
            kind = 'available'
            exclude_repos = '*'
            include_repos = filters.pop('repoid')
        pkglist = self._handle_get_package_list(kind, allow_duplicates, False,
                include_repos=include_repos, exclude_repos=exclude_repos,
                transform=False)
        matches = _get_package_filter_function(filters, exact_match)
        result = [p for p in pkglist if matches(p)]
        if sort is True:
            result.sort()
        LOG.debug("%d packages matching", len(result))
        if transform is True:
            # caching has been already done by _handle_get_package_list()
            result = self._transform_packages(result, cache_packages=False)
        return result

    @_needs_database
    def _handle_install_package(self, pkg, force=False):
        """
        Handler for package installation job.
        @return installed package instance
        """
        if isinstance(pkg, basestring):
            pkgs = self._handle_filter_packages(
                    'available' if force else 'avail_notinst',
                    allow_duplicates=False, sort=True,
                    transform=False, nevra=pkg)
            if len(pkgs) < 1:
                raise errors.PackageNotFound('No available package matches'
                        ' nevra "%s".' % pkg)
            elif len(pkgs) > 1:
                LOG.warn('multiple packages matches nevra "%s": [%s]',
                        pkg, ", ".join(p.nevra for p in pkgs))
            pkg_desired = pkgs[-1]
        else:
            pkg_desired = self._lookup_package(pkg)
        if is_pkg_installed(pkg_desired, self._yum_base.rpmdb):
            if force is False:
                raise errors.PackageAlreadyInstalled(pkg)
            action = "reinstall"
        else:
            action = "install"
        getattr(self._yum_base, action)(pkg_desired)
        self._run_transaction(action)
        installed = self._handle_filter_packages("installed", False, False,
                nevra=util.pkg2nevra(pkg_desired, with_epoch="ALWAYS"))
        if len(installed) < 1:
            raise errors.TransactionExecutionFailed(
                    "Failed to install desired package %s." % pkg)
        return installed[0]

    @_needs_database
    def _handle_remove_package(self, pkg):
        """
        Handler for package removal job.
        """
        if isinstance(pkg, basestring):
            pkgs = self._handle_filter_packages('installed',
                    allow_duplicates=False, sort=False,
                    transform=False, nevra=pkg)
            if len(pkgs) < 1:
                raise errors.PackageNotFound('No available package matches'
                        ' nevra "%s".' % pkg)
            pkg = pkgs[-1]
        else:
            pkg = self._lookup_package(pkg)
        self._yum_base.remove(pkg)
        self._run_transaction("remove")

    @_needs_database
    def _handle_update_to_package(self, pkg):
        """
        Handler for specific package update job.
        @return package corresponding to pkg after update
        """
        if isinstance(pkg, basestring):
            pkgs = self._handle_filter_packages('available',
                    allow_duplicates=False, sort=False,
                    transform=False, nevra=pkg)
            if len(pkgs) < 1:
                raise errors.PackageNotFound('No available package matches'
                        ' nevra "%s".' % pkg)
            pkg_desired = pkgs[-1]
        else:
            pkg_desired = self._lookup_package(pkg)
        self._yum_base.update(update_to=True,
                name=pkg_desired.name,
                epoch=pkg_desired.epoch,
                version=pkg_desired.version,
                release=pkg_desired.release,
                arch=pkg_desired.arch)
        self._run_transaction("update")
        installed = self._handle_filter_packages("installed", False, False,
                **pkg.key_props)
        if len(installed) < 1:
            raise errors.TransactionExecutionFailed(
                    "Failed to update to desired package %s." % pkg)
        return installed[0]

    @_needs_database
    def _handle_update_package(self, pkg, to_epoch, to_version, to_release,
            force=False):
        """
        Handler for package update job.
        @return updated package instance
        """
        if isinstance(pkg, basestring):
            pkgs = self._handle_filter_packages('installed',
                    allow_duplicates=False, sort=False,
                    transform=False, nevra=pkg)
            if len(pkgs) < 1:
                raise errors.PackageNotFound('No available package matches'
                        ' nevra "%s".' % pkg)
            pkg = pkgs[-1]
        else:
            pkg = self._lookup_package(pkg)
        if not is_pkg_installed(pkg, self._yum_base.rpmdb):
            raise errors.PackageNotInstalled(pkg)
        kwargs = { "name" : pkg.name, "arch" : pkg.arch }
        if any(v is not None for v in (to_epoch, to_version, to_release)):
            kwargs["update_to"] = True
        if to_epoch:
            kwargs["to_epoch"] = to_epoch
        if to_version:
            kwargs["to_version"] = to_version
        if to_release:
            kwargs["to_release"] = to_release
        self._yum_base.update(**kwargs)
        self._run_transaction("update")
        kwargs = dict(  (k[3:] if k.startswith("to_") else k, v)
                     for k, v in kwargs.items())
        installed = self._handle_filter_packages(
                "installed", False, False, **kwargs)
        if len(installed) < 1:
            raise errors.TransactionExecutionFailed(
                    "Failed to update package %s." % pkg)
        return installed[0]

    @_needs_database
    def _handle_check_package(self, pkg, file_name=None):
        """
        @return PackageCheck instance for requested package
        """
        if isinstance(pkg, basestring):
            pkgs = self._handle_filter_packages('installed',
                    allow_duplicates=False, sort=False,
                    transform=False, nevra=pkg)
            if len(pkgs) < 1:
                raise errors.PackageNotFound('No available package matches'
                        ' nevra "%s".' % pkg)
            rpm = pkgs[-1]
            pkg = self._transform_packages((rpm, ), cache_packages=False)[0]
        else:
            rpm = self._lookup_package(pkg)
        if not is_pkg_installed(rpm, self._yum_base.rpmdb):
            raise errors.PackageNotInstalled(rpm)
        vpkg = yum.packages._RPMVerifyPackage(rpm, rpm.hdr.fiFromHeader(),
                packagecheck.pkg_checksum_type(rpm), [], True)
        return (pkg, packagecheck.make_package_check_from_db(vpkg,
                file_name=file_name))

    @_needs_database
    def _handle_check_package_file(self, pkg, file_name):
        """
        @return PackageCheck instance for requested package containing
            just one PackageFile instance for given ``file_name``.
        """
        return self._handle_check_package(pkg, file_name)

    @_needs_database
    def _handle_install_package_from_uri(self, uri,
            update_only=False, force=False):
        """
        @return installed PackageInfo instance
        """
        try:
            pkg = yum.packages.YumUrlPackage(self._yum_base,
                    ts=self._yum_base.rpmdb.readOnlyTS(), url=uri,
                    ua=default_grabber.opts.user_agent)
        except yum.Errors.MiscError as exc:
            raise errors.PackageOpenError(uri, str(exc))
        installed = self._handle_filter_packages("installed", False, False,
                nevra=util.pkg2nevra(pkg, with_epoch="ALWAYS"))
        if installed and force is False:
            raise errors.PackageAlreadyInstalled(pkg)
        kwargs = { 'po' : pkg }
        if installed:
            action = 'reinstallLocal'
        else:
            action = 'installLocal'
            kwargs = { 'updateonly' : update_only }
        getattr(self._yum_base, action)(uri, **kwargs)
        self._run_transaction('installLocal')
        installed = self._handle_filter_packages("installed", False, False,
                nevra=util.pkg2nevra(pkg, with_epoch="ALWAYS"))
        if len(installed) < 1:
            raise errors.TransactionExecutionFailed(
                    "Failed to install desired package %s." % pkg)
        return installed[0]

    @_needs_database
    def _handle_get_repository_list(self, kind, transform=True):
        """
        @return list of yumdb.Repository instances
        """
        if kind == 'enabled':
            repos = sorted(self._yum_base.repos.listEnabled())
        else:
            repos = self._yum_base.repos.repos.values()
            if kind == 'disabled':
                repos = [repo for repo in repos if not repo.enabled]
            repos.sort()
        if transform:
            repos = [repository.make_repository_from_db(r) for r in repos]
        LOG.debug("returning %d repositories from %s",
                len(repos), kind)
        return repos

    @_needs_database
    def _handle_filter_repositories(self, kind, **filters):
        """
        @return list of yumdb.Repository instances -- filtered
        """
        filters = dict((k, v) for k, v in filters.items() if v is not None)
        if 'repoid' in filters:
            try:
                repo = repository.make_repository_from_db(
                            self._yum_base.repos.getRepo(filters["repoid"]))
                if (  (kind == "enabled"  and not repo.enabled)
                   or (kind == "disabled" and repo.enabled)):
                    LOG.warn(
                            'no such repository with id="%s"matching filters',
                            filters['repoid'])
                    return []
                LOG.debug(
                        "exactly one repository matching filters found")
                return [repo]
            except (KeyError, yum.Errors.RepoError):
                LOG.warn('repository with id="%s" could not be found',
                        filters['repoid'])
                raise errors.RepositoryNotFound(filters['repoid'])

        repos = self._handle_get_repository_list(kind, transform=False)
        result = []
        for repo in repos:
            # do the filtering and safe transformed repo into result
            for prop, value in filters.items():
                if repository.get_prop_from_yum_repo(repo, prop) != value:
                    # did not pass the filter
                    break
            else: # all properties passed
                result.append(repository.make_repository_from_db(repo))
        LOG.debug("found %d repositories matching", len(result))
        return result

    @_needs_database
    def _handle_set_repository_enabled(self, repo, enable):
        """
        @return previous enabled state
        """
        if isinstance(repo, repository.Repository):
            repoid = repo.repoid
        else:
            repoid = repo
        try:
            repo = self._yum_base.repos.getRepo(repoid)
        except (KeyError, yum.Errors.RepoError):
            raise errors.RepositoryNotFound(repoid)
        res = repo.enabled
        try:
            if enable ^ res:
                if enable is True:
                    LOG.info("enabling repository %s", repo)
                    repo.enable()
                else:
                    LOG.info("disabling repository %s", repo)
                    repo.disable()
                try:
                    yum.config.writeRawRepoFile(repo, only=["enabled"])
                except Exception as exc:
                    raise errors.RepositoryChangeError(
                            'failed to modify repository "%s": %s - %s' % (
                                repo, exc.__class__.__name__, str(exc)))
            else:
                LOG.info("no change for repo %s", repo)
        except yum.Errors.RepoError as exc:
            raise errors.RepositoryChangeError(
                    'failed to modify repository "%s": %s' % (repo, str(exc)))
        return res

    # *************************************************************************
    # Public properties
    # *************************************************************************
    @property
    def uplink(self):
        """
        @return input queue for jobs
        """
        return self._queue_in

    @property
    def downlink(self):
        """
        @return output queue for job results
        """
        return self._queue_out

    # *************************************************************************
    # Public methods
    # *************************************************************************
    def run(self):
        """
        Process's entry point. After initial setup it calls _main_loop().
        """
        setup_logging()
        global LOG
        LOG = logging.getLogger(__name__)
        LOG.info("running as pid=%d", self.pid)
        self._pkg_cache = weakref.WeakValueDictionary()

        self._main_loop()
        LOG.info("terminating")

